import pytest

from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.modules.datasets_client import DatasetsClient
from cnvrgv2.proxy import Proxy
from cnvrgv2.utils.json_api_format import JAF


class TestDatasetsClient:
    def test_init_no_credentials(self):
        with pytest.raises(CnvrgArgumentsError):
            DatasetsClient(None)

    def test_init_valid_credentials(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        assert client._context
        assert client._route

    def test_get_with_slug(self, mocker, empty_context, empty_cnvrg):
        slug = "test-slug"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "dataset",
                    "attributes": {
                        "title": "test-title",
                        "slug": slug,
                    }
                }
            }
            return JAF(response=response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        model = client.get(slug)

        # The function will not be called on init since its lazy-loading
        assert mocked_api_call.call_count == 0

        # The function will not be called on init since its lazy-loading
        assert model.title == "test-title"
        assert mocked_api_call.call_count > 0

        assert type(model) is Dataset

    def test_get_with_empty_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get("")

    def test_get_with_none_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get(None)

    def test_get_with_wrong_type_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get({"say": "whaaat??"})

    def test_valid_create(self, mocker, empty_context, empty_cnvrg):
        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "dataset",
                    "attributes": {
                        "title": "test-title",
                        "slug": "test-slug",
                    }
                }
            }
            return JAF(response=response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        model_no_cat = client.create("test-title")
        model_with_cat = client.create("test-title", category="images")

        # TODO: add attribute checks. added call_count until decided what to check
        assert type(model_no_cat) is Dataset
        assert type(model_with_cat) is Dataset
        assert mocked_api_call.call_count > 0

    def test_create_with_empty_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create("")

    def test_create_with_none_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create(None)

    def test_create_with_wrong_type_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create({"say": "whaaat??"})

    def test_create_with_empty_category(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create("name", category="")

    def test_create_with_none_category(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create("name", category=None)

    def test_create_with_wrong_type_category(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create("name", category={"say": "whaaat??"})

    def test_create_with_wrong_category_value(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = DatasetsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create("name", category="fake")

    # TODO write this test for CLI
    # def test_verify_datasets(self, mocker, empty_context,empty_cnvrg):
    #     empty_cnvrg._context = empty_context
    #     client = DatasetsClient(empty_cnvrg)
    #     mocker.patch.object(time, "sleep", return_value=None)
    #
    #     with mocker.patch.object(Dataset, "verify", return_value=True):
    #         assert client.verify_datasets(['test1', 'test2'])
    #
    #     with mocker.patch.object(Dataset, "verify", side_effect=[False, False, True, True]):
    #         assert client.verify_datasets(['test1', 'test2'])
    #
    #     with mocker.patch.object(Dataset, "verify", side_effect=[False, False, False, False, True, True]):
    #         assert client.verify_datasets(['test1', 'test2'])
    #
    #     with mocker.patch.object(Dataset, "verify", side_effect=[False, False, False, False]):
    #         mocker.patch.object(DatasetsClient, "seconds_diff", side_effect=[0, 1, 2])
    #         assert client.verify_datasets(['test1', 'test2'], 2) is False
    #
    #     x = mocker.patch.object(Dataset, "verify", side_effect=[False, False, True, False])
    #     mocker.patch.object(DatasetsClient, "seconds_diff", side_effect=[0, 1, 2])
    #     assert client.verify_datasets(['test1', 'test2'], 2) is False
    #     assert x.call_count == 4
    #
    #     with mocker.patch.object(Dataset, "verify", side_effect=[False, False, True, True]):
    #         mocker.patch.object(DatasetsClient, "seconds_diff", side_effect=[0, 1, 2])
    #         assert client.verify_datasets(['test1', 'test2'], 2) is True



import mimetypes
import os

import google.auth.transport.requests as tr_requests
from google.oauth2 import service_account
from google.resumable_media.requests import ChunkedDownload, ResumableUpload

from cnvrgv2.config import CONFIG_FOLDER_NAME
from cnvrgv2.data.clients.base_storage_client import BaseStorageClient
from cnvrgv2.utils.retry import retry
from cnvrgv2.utils.storage_utils import create_dir_if_not_exists, download_file


class GCPStorage(BaseStorageClient):
    def __init__(self, storage_meta):
        super().__init__(storage_meta)
        self.chunk_size = 262144 * 5  # needs to be a multiply of 256 kb. 262144 bytes = 256 kb

        self.key_file = os.path.join(os.path.expanduser("~"), CONFIG_FOLDER_NAME, ".gcp_cred.json")

        self.props = self._decrypt_dict(storage_meta, keys=["credentials", "bucket_name", "project_id"])

        download_file(self.props["credentials"], self.key_file)

        self.download_url_template = 'https://storage.googleapis.com/{bucket_name}/{object_path}'
        self.upload_url = 'https://www.googleapis.com/upload/storage/v1/b/{bucket_name}/o/?uploadType=resumable'
        self.credential_scope = 'https://www.googleapis.com/auth/devstorage.read_write'
        self.transport = self._get_transport()

    @retry(log_error=True)
    def upload_single_file(self, local_path, object_path, progress_bar):
        try:
            progress_incrementer = self.progress_callback(progress_bar)
            metadata = {'name': object_path}
            content_type = mimetypes.guess_type(local_path, strict=True)[0] or "plain/text"
            media_url = self.upload_url.format(object_path=object_path, bucket_name=self.props['bucket_name'])

            with open(local_path, "rb") as file_stream:
                # Creates a ResumableUpload object that allows to transmit data in chunks.
                # Needs to be followed by an initiate command, that makes the first request to gcp, and gets
                # essential information for the uploading process to work
                upload = ResumableUpload(media_url, self.chunk_size)
                upload.initiate(self.transport, file_stream, metadata, content_type)
                while not upload.finished:
                    last_bytes_uploaded = upload.bytes_uploaded
                    upload.transmit_next_chunk(self.transport)
                    last_bytes_uploaded = upload.bytes_uploaded - last_bytes_uploaded
                    progress_incrementer(last_bytes_uploaded)

        except Exception as e:
            print(e)

    @retry(log_error=True)
    def download_single_file(self, local_path, object_path, progress_bar=None):
        try:
            create_dir_if_not_exists(local_path)
            if not object_path:
                return

            progress_incrementer = self.progress_callback(progress_bar)
            media_url = self.download_url_template.format(
                object_path=object_path,
                bucket_name=self.props['bucket_name']
            )
            with open(local_path, "wb") as file_stream:
                download = ChunkedDownload(media_url, self.chunk_size, file_stream)
                while not download.finished:
                    response = download.consume_next_chunk(self.transport)
                    progress_incrementer(len(response.content))

        except Exception as e:
            print(e)

    def _get_transport(self):
        credentials = service_account.Credentials.from_service_account_file(self.key_file)
        scoped_credentials = credentials.with_scopes([self.credential_scope])
        return tr_requests.AuthorizedSession(scoped_credentials)



from cnvrgv2.errors import CnvrgHttpError, CnvrgArgumentsError
from cnvrgv2.modules.datasource import Datasource, StorageTypes
from cnvrgv2.config import GLOBAL_CNVRG_PATH, error_messages
from pathlib import Path
import os
from dotenv import load_dotenv
import pytest
from tests.e2e.conftest import call_database
from tests.long_running_tests.datasources.conftest import datasource_test_bukcet_name, datasource_test_region
from cnvrgv2.cnvrg import Cnvrg


class TestDatasources:
    storage_type = StorageTypes.S3
    storage_bucket_name = None
    storage_endpoint = None
    storage_region = None
    storage_path = None
    storage_credentials_user_rw = None
    storage_credentials_user_ro = None
    datasource_rw = None
    datasource_ro = None
    config_env_vars = ['DATASOURCE_TEST_BUCKET_NAME', 'DATASOURCE_TEST_REGION',
                       'DATASOURCE_TEST_USER_RW_ACCESS_KEY_ID', 'DATASOURCE_TEST_USER_RW_SECRET_ACCESS_KEY',
                       'DATASOURCE_TEST_USER_RO_ACCESS_KEY_ID', 'DATASOURCE_TEST_USER_RO_SECRET_ACCESS_KEY']

    @classmethod
    def set_datasource_aws_config(cls):
        # Path to the AWS config file
        config_file = os.path.join(GLOBAL_CNVRG_PATH, '.env')

        # load from config file if running locally.
        # Env vars should exist when running through github actions
        if Path(config_file).is_file():
            load_dotenv(config_file)

        bucket_name = datasource_test_bukcet_name
        region = datasource_test_region
        user1_access_key_id = os.getenv(key='DATASOURCE_TEST_USER_RW_ACCESS_KEY_ID')
        user1_secret_access_key = os.getenv(key='DATASOURCE_TEST_USER_RW_SECRET_ACCESS_KEY')
        user2_access_key_id = os.getenv(key='DATASOURCE_TEST_USER_RO_ACCESS_KEY_ID')
        user2_secret_access_key = os.getenv(key='DATASOURCE_TEST_USER_RO_SECRET_ACCESS_KEY')

        cls.storage_bucket_name = bucket_name
        cls.storage_region = region
        cls.storage_credentials_user_rw = {
            "access_key_id": user1_access_key_id,
            "secret_access_key": user1_secret_access_key
        }
        cls.storage_credentials_user_ro = {
            "access_key_id": user2_access_key_id,
            "secret_access_key": user2_secret_access_key
        }

    @classmethod
    def create_datasources(cls, class_context, e2e_client):
        path = class_context.generate_name(5)
        cls.storage_path = path + '/'
        cls.datasource_rw = e2e_client.datasources.create(name=class_context.generate_name(5),
                                                          storage_type=cls.storage_type,
                                                          path=path,
                                                          bucket_name=cls.storage_bucket_name,
                                                          region=cls.storage_region,
                                                          credentials=cls.storage_credentials_user_rw)

        cls.datasource_ro = e2e_client.datasources.create(name=class_context.generate_name(5),
                                                          storage_type=cls.storage_type,
                                                          path=path,
                                                          bucket_name=cls.storage_bucket_name,
                                                          region=cls.storage_region,
                                                          credentials=cls.storage_credentials_user_ro)

    @classmethod
    def clear_env(cls):
        for variable_name in cls.config_env_vars:
            if variable_name in os.environ:
                os.environ.pop(variable_name)

    @classmethod
    def clear_bucket(cls):
        cls.datasource_rw.remove_file(cls.datasource_rw.path)

    @classmethod
    @pytest.fixture(scope='class', autouse=True)
    def setup_and_teardown_class(cls, class_context, e2e_client):
        # Setup Class - Executed once per test class
        cls.set_datasource_aws_config()
        # Skip tests if config is missing
        if any(var is None for var in [cls.storage_bucket_name, cls.storage_region,
                                       cls.storage_credentials_user_rw['access_key_id'],
                                       cls.storage_credentials_user_rw['secret_access_key'],
                                       cls.storage_credentials_user_ro['access_key_id'],
                                       cls.storage_credentials_user_ro['secret_access_key']]):
            pytest.skip("Skipping tests due to CONFIGURATION MISSING")
        cls.create_datasources(class_context, e2e_client)
        yield
        # Teardown Class - Executed once after all test methods in the class
        cls.clear_env()

    @staticmethod
    def cleanup(context_prefix):
        # Clean Data Sources by name, will clean all data sources where name starts with the class prefix
        # Create name with : class_context.generate_name(5)
        delete_user_command = "DELETE FROM datasources WHERE name LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    # region Create
    def test_create_private(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        count_before = e2e_client.datasources.list_count(sort="id")
        new_datasource = self.create_random_datasource(e2e_client, name, credentials=self.storage_credentials_user_ro)
        count_after = e2e_client.datasources.list_count(sort="id")

        datasource = e2e_client.datasources.get(slug=new_datasource.slug)
        assert datasource is not None
        assert datasource.slug == new_datasource.slug
        assert count_after is count_before + 1

        assert type(datasource) == Datasource
        assert datasource.name == name
        assert datasource.storage_type == self.storage_type.value
        assert datasource.bucket_name == self.storage_bucket_name
        assert datasource.path == self.storage_path
        assert datasource.endpoint == self.storage_endpoint
        assert datasource.region == self.storage_region
        assert datasource.public is False

    def test_create_public(self, class_context, e2e_client):
        # We create names with class prefix so that the cleanup will find them.
        name = class_context.generate_name(5)
        count_before = e2e_client.datasources.list_count(sort="id")
        new_datasource = self.create_random_datasource(e2e_client, name, public=True,
                                                       credentials=self.storage_credentials_user_ro)
        count_after = e2e_client.datasources.list_count(sort="id")

        datasource = e2e_client.datasources.get(slug=new_datasource.slug)
        assert datasource is not None
        assert datasource.slug == new_datasource.slug
        assert count_after is count_before + 1

        assert type(datasource) == Datasource
        assert datasource.name == name
        assert datasource.storage_type == self.storage_type.value
        assert datasource.bucket_name == self.storage_bucket_name
        assert datasource.path == self.storage_path
        assert datasource.endpoint == self.storage_endpoint
        assert datasource.region == self.storage_region
        assert datasource.public is True

    def test_create_unauthorized_for_non_admin_user(self, domain, class_context, e2e_client, e2e_data_scientist_user):
        # We create names with class prefix so that the cleanup will find them.
        name = class_context.generate_name(5)
        client_of_non_admin_user = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
            organization=e2e_client._context.organization
        )
        with pytest.raises(CnvrgHttpError) as exception_info:
            client_of_non_admin_user.datasources.create(name=name,
                                                        storage_type=self.storage_type,
                                                        bucket_name=self.storage_bucket_name,
                                                        region=self.storage_region,
                                                        credentials=self.storage_credentials_user_ro)
        assert 'Unauthorized' in str(exception_info.value)

    def test_create_missing_mandatory_field(self, e2e_client):
        with pytest.raises(TypeError) as exception_info:
            e2e_client.datasources.create()
        assert "required positional argument" in str(exception_info.value)
        assert "name" in str(exception_info.value)
        assert "storage_type" in str(exception_info.value)
        assert "bucket_name" in str(exception_info.value)
        assert "credentials" in str(exception_info.value)
        assert "region" in str(exception_info.value)

    def test_create_faulty_params(self, class_context, e2e_client):
        # Faulty name (not str)
        # sdk validation error
        with pytest.raises(CnvrgArgumentsError) as exception_info:
            self.create_random_datasource(e2e_client,
                                          name=123)
        assert "name" in str(exception_info.value)
        assert "not a valid" in str(exception_info.value)

        # Faulty name (special chars)
        # server validation error
        # with pytest.raises(CnvrgHttpError) as exception_info:
        #     e2e_client.datasources.create(storage_type=self.storage_type,
        #                                   name="test@#$%^&",
        #                                   bucket_name=self.storage_bucket_name,
        #                                   path=self.storage_path,
        #                                   region=self.storage_region,
        #                                   credentials=self.storage_credentials_user_rw)
        # assert "Invalid Input" in str(exception_info.value.args)
        # assert exception_info.value.status_code == unprocessable_entity_status_code
        # Faulty bucket name (not str)
        with pytest.raises(CnvrgArgumentsError) as exception_info:
            e2e_client.datasources.create(storage_type=self.storage_type,
                                          name=class_context.generate_name(5),
                                          bucket_name=123,
                                          path=self.storage_path,
                                          region=self.storage_region,
                                          credentials=self.storage_credentials_user_rw)

    def test_create_fails_for_duplicate_name(self, class_context, e2e_client):
        # We create names with class prefix so that the cleanup will find them.
        name = class_context.generate_name(5)
        e2e_client.datasources.create(name=name,
                                      storage_type=self.storage_type,
                                      bucket_name=self.storage_bucket_name,
                                      region=self.storage_region,
                                      credentials=self.storage_credentials_user_rw)
        with pytest.raises(CnvrgHttpError) as exception_info:
            e2e_client.datasources.create(name=name,
                                          storage_type=self.storage_type,
                                          bucket_name=self.storage_bucket_name,
                                          path=self.storage_path,
                                          region=self.storage_region,
                                          credentials=self.storage_credentials_user_rw)
        assert "already exists" in str(exception_info.value)

    # region test validity of users

    def test_create_and_assign_not_authorized_collaborator_should_throw_error_and_not_create_datasource(self,
                                                                                                        class_context,
                                                                                                        e2e_client,
                                                                                                        e2e_user,
                                                                                                        e2e_temp_user):

        # Get count (as admin)
        count_before = e2e_client.datasources.list_count(sort="id")

        # Creating a data source with unauthorized users should throw an error with emails of unauthorized users only.
        # And not create a new data source
        with pytest.raises(CnvrgHttpError) as exception_info:
            self.create_random_datasource(e2e_client,
                                          name=class_context.generate_name(5),
                                          collaborators=['manager_role17@test.com',
                                                         'fake@user.com',
                                                         e2e_temp_user['email']],
                                          credentials=self.storage_credentials_user_rw)

        assert 'manager_role17@test.com, fake@user.com' in str(exception_info.value)
        assert e2e_user['email'] not in str(exception_info.value)

        assert e2e_client.datasources.list_count(sort="id") is count_before

    def test_create_and_assign_authorized_collaborator(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        datasource = self.create_random_datasource(e2e_client,
                                                   name=class_context.generate_name(5),
                                                   collaborators=[e2e_temp_user['email']],
                                                   credentials=self.storage_credentials_user_rw)

        # Collaborators should include the user who created the data source + users params (e2e_temp_user)
        assert len(datasource.collaborators) == 2
        assert e2e_user['email'] in datasource.collaborators
        assert e2e_temp_user['email'] in datasource.collaborators

    def test_create_no_collaborators(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        datasource = self.create_random_datasource(e2e_client,
                                                   name=class_context.generate_name(5),
                                                   credentials=self.storage_credentials_user_rw)

        # Collaborators should include the user who created the data source only.
        assert len(datasource.collaborators) == 1
        assert e2e_user['email'] in datasource.collaborators

    def test_create_and_assign_admin_as_collaborator(self, class_context, e2e_client, e2e_user):
        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        datasource = self.create_random_datasource(e2e_client,
                                                   name=class_context.generate_name(5),
                                                   collaborators=[e2e_user['email']],
                                                   credentials=self.storage_credentials_user_rw)

        # Datasource collaborators = admin + users.
        # e2e_user is the owner of the datasource. It should not be added as user too.
        # Meaning it should appear once in the list of collaborators.
        assert len(datasource.collaborators) == 1
        assert e2e_user['email'] in datasource.collaborators

    # endregion

    # region test validity of credentials
    def test_credentials_format(self, class_context, e2e_client):
        # We create names with class prefix so that the cleanup will find them.
        name = class_context.generate_name(5)

        storage_credentials = {"access_key": "aaa", "secret_access": "bbb"}
        with pytest.raises(CnvrgHttpError) as exception_info:
            e2e_client.datasources.create(name=name,
                                          storage_type=self.storage_type,
                                          region=self.storage_region,
                                          bucket_name=self.storage_bucket_name,
                                          credentials=storage_credentials,
                                          )

        assert "Credentials not valid" in str(exception_info.value)

        storage_credentials = {"access_key_id": "aaa"}
        with pytest.raises(CnvrgHttpError) as exception_info:
            e2e_client.datasources.create(name=name,
                                          storage_type=self.storage_type,
                                          region=self.storage_region,
                                          bucket_name=self.storage_bucket_name,
                                          credentials=storage_credentials)

        assert "Credentials not valid" in str(exception_info.value)

    # endregion

    # endregion

    # region Get

    def test_get(self, class_context, e2e_client):
        # We create names with class prefix so that the cleanup will find them.
        name = class_context.generate_name(5)

        datasource = self.create_random_datasource(e2e_client,
                                                   name,
                                                   credentials=self.storage_credentials_user_ro)

        datasource = e2e_client.datasources.get(slug=datasource.slug)
        assert type(datasource) == Datasource
        assert datasource.name == name
        assert datasource.storage_type == self.storage_type.value
        assert datasource.bucket_name == self.storage_bucket_name
        assert datasource.path == self.storage_path

    def test_get_faulty_slug(self, class_context, e2e_client):
        with pytest.raises(CnvrgArgumentsError) as exception_info:
            e2e_client.datasources.get(slug=123)
        assert error_messages.DATASOURCE_GET_FAULTY_SLUG in str(exception_info.value)

        with pytest.raises(CnvrgArgumentsError) as exception_info:
            e2e_client.datasources.get(slug=None)
        assert error_messages.DATASOURCE_GET_FAULTY_SLUG in str(exception_info.value)

    def test_get_non_existent(self, class_context, e2e_client):
        # We create names with class prefix so that the cleanup will find them.
        name = class_context.generate_name(5)

        with pytest.raises(CnvrgHttpError) as exception_info:
            e2e_client.datasources.get(slug=name)
        assert "The requested resource could not be found" in str(exception_info.value)

    def test_get_unauthorized_user(self, domain, class_context, e2e_client, e2e_data_scientist_user):
        new_datasource = self.create_random_datasource(e2e_client,
                                                       name=class_context.generate_name(5),
                                                       credentials=self.storage_credentials_user_rw)
        client_of_non_admin_user = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
            organization=e2e_client._context.organization
        )
        slug = new_datasource.slug
        with pytest.raises(CnvrgHttpError):
            ds = client_of_non_admin_user.datasources.get(slug=slug)
            # without reload, it doesn't perform get
            ds.reload()

    def test_get_authorized_user(self, domain, class_context, e2e_client, e2e_data_scientist_user):
        new_datasource = self.create_random_datasource(e2e_client,
                                                       name=class_context.generate_name(5),
                                                       credentials=self.storage_credentials_user_rw,
                                                       collaborators=[e2e_data_scientist_user['email']])
        client_of_non_admin_user = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
            organization=e2e_client._context.organization
        )
        slug = new_datasource.slug
        ds = client_of_non_admin_user.datasources.get(slug=slug)
        # without reload, it doesn't perform get
        ds.reload()
        assert slug == ds.slug

    def test_get_all_as_admin(self, domain, class_context, e2e_client, e2e_data_scientist_user):
        new_datasource = self.create_random_datasource(e2e_client,
                                                       name=class_context.generate_name(5),
                                                       credentials=self.storage_credentials_user_rw,
                                                       collaborators=[e2e_data_scientist_user['email']])
        client_of_non_admin_user = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
            organization=e2e_client._context.organization
        )
        slug = new_datasource.slug
        client_of_non_admin_user.datasources.get(slug=slug)
        # without reload, it doesn't perform get
        datasource = e2e_client.datasources.get(slug=new_datasource.slug)
        datasource.reload()
        assert slug == datasource.slug

    def test_get_as_admin(self, class_context, e2e_client):
        new_datasource = self.create_random_datasource(e2e_client,
                                                       name=class_context.generate_name(5),
                                                       credentials=self.storage_credentials_user_rw)
        datasource = e2e_client.datasources.get(slug=new_datasource.slug)
        datasource.reload()
        assert new_datasource.slug == datasource.slug

    # endregion

    # region List
    def test_list(self, class_context, e2e_client):
        # Clean the db before testing list
        TestDatasources.cleanup(class_context.prefix)
        assert e2e_client.datasources.list_count(sort="id") == 0

        # We create names with class prefix so that the cleanup will find them.
        name_base = class_context.generate_name(5)
        for i in range(40):
            name = "{}{}".format(name_base, i)
            self.create_random_datasource(e2e_client,
                                          name,
                                          credentials=self.storage_credentials_user_rw)

        # Check descending order works
        idx = 39
        datasources = e2e_client.datasources.list()
        for datasource in datasources:
            assert datasource.name == "{}{}".format(name_base, idx)
            idx -= 1

        # Check ascending order works
        idx = 0
        datasources = e2e_client.datasources.list(sort="id")
        for datasource in datasources:
            assert datasource.name == "{}{}".format(name_base, idx)
            idx += 1

    def test_list_filter(self, class_context, e2e_client):
        # Clean the db before testing list
        TestDatasources.cleanup(class_context.prefix)
        assert e2e_client.datasources.list_count(sort="id") == 0
        total = 40
        # We create names with class prefix so that the cleanup will find them.
        name_base = class_context.generate_name(5)
        for i in range(total):
            name = "{}{}".format(name_base, i)
            self.create_random_datasource(e2e_client,
                                          name,
                                          credentials=self.storage_credentials_user_rw)

        idx = total - 1  # count starts at 0
        s3_filter = {'operator': "AND",
                     'conditions': [{'key': "type", 'operator': "is", 'value': StorageTypes.S3.value}]}
        datasources = e2e_client.datasources.list(filter=s3_filter)
        assert e2e_client.datasources.list_count(sort="id", filter=s3_filter) == total
        for datasource in datasources:
            assert datasource.name == "{}{}".format(name_base, idx)
            assert datasource.storage_type == StorageTypes.S3.value
            idx -= 1

        minio_filter = {'operator': "AND",
                        'conditions': [{'key': "type", 'operator': "is", 'value': StorageTypes.MINIO.value}]}
        assert e2e_client.datasources.list_count(filter=minio_filter) == 0

        fake_filter = {'operator': "AND",
                       'conditions': [{'key': "type", 'operator': "is", 'value': "fake"}]}
        assert e2e_client.datasources.list_count(filter=fake_filter) == 0

    def test_list_assigned_to_user_only(self, class_context, e2e_client, e2e_user, e2e_temp_user, domain):
        # Clean the db before testing list
        TestDatasources.cleanup(class_context.prefix)
        assert e2e_client.datasources.list_count(sort="id") == 0

        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        bucket_assigned_to_temp_user = self.create_random_datasource(e2e_client,
                                                                     name=class_context.generate_name(5),
                                                                     collaborators=[e2e_temp_user['email']],
                                                                     credentials=self.storage_credentials_user_rw)

        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        self.create_random_datasource(e2e_client,
                                      name=class_context.generate_name(5),
                                      credentials=self.storage_credentials_user_rw)

        # List as temp_user (not admin), it should only list data sources assigned to temp_user.
        temp_user_cnvrg = Cnvrg(domain=domain, email=e2e_temp_user["email"], password=e2e_temp_user["password"])

        datasources_count = temp_user_cnvrg.datasources.list_count(sort="id")
        assert datasources_count == 1
        for ds in temp_user_cnvrg.datasources.list():
            assert len(ds.collaborators) >= 2  # at least temp_user & admin (it could be assigned to more collaborators too)
            assert e2e_user['email'] in ds.collaborators
            assert e2e_temp_user['email'] in ds.collaborators
            assert ds.slug == bucket_assigned_to_temp_user.slug

    def test_list_assigned_to_user_only_count_0(self, class_context, e2e_client, e2e_user, e2e_temp_user, domain):
        # Clean the db before testing list
        TestDatasources.cleanup(class_context.prefix)
        assert e2e_client.datasources.list_count(sort="id") == 0

        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        # no collaborators assigned to the data source
        self.create_random_datasource(e2e_client,
                                      name=class_context.generate_name(5),
                                      credentials=self.storage_credentials_user_rw)

        # list as temp_user
        temp_user_cnvrg = Cnvrg(domain=domain, email=e2e_temp_user["email"], password=e2e_temp_user["password"])
        datasources_count = temp_user_cnvrg.datasources.list_count(sort="id")
        assert datasources_count == 0

    def test_list_all_as_admin(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        # Clean the db before testing list
        TestDatasources.cleanup(class_context.prefix)
        assert e2e_client.datasources.list_count(sort="id") == 0
        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        bucket_assigned_to_temp_user = self.create_random_datasource(e2e_client,
                                                                     name=class_context.generate_name(5),
                                                                     collaborators=[e2e_temp_user['email']],
                                                                     credentials=self.storage_credentials_user_rw)

        # We create names with class prefix so that the cleanup will find them.
        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        bucket_not_assigned_to_collaborators = self.create_random_datasource(e2e_client,
                                                                             name=class_context.generate_name(5),
                                                                             credentials=self.storage_credentials_user_rw)

        # list as admin

        datasources_count = e2e_client.datasources.list_count(sort="id")
        assert datasources_count == 2
        for ds in e2e_client.datasources.list():
            assert e2e_user['email'] in ds.collaborators
            assert ds.slug in [bucket_assigned_to_temp_user.slug, bucket_not_assigned_to_collaborators.slug]

    def test_list_public(self, class_context, e2e_client, e2e_user, e2e_temp_user, domain):
        # Clean the db before testing list
        TestDatasources.cleanup(class_context.prefix)
        assert e2e_client.datasources.list_count(sort="id") == 0

        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        # no collaborators assigned to the data source
        self.create_random_datasource(e2e_client,
                                      name=class_context.generate_name(5),
                                      credentials=self.storage_credentials_user_rw)

        # list as temp_user
        temp_user_cnvrg = Cnvrg(domain=domain, email=e2e_temp_user["email"], password=e2e_temp_user["password"])
        datasources_count = temp_user_cnvrg.datasources.list_count(sort="id")
        assert datasources_count == 0
        # Clean the db before testing list
        TestDatasources.cleanup(class_context.prefix)
        assert e2e_client.datasources.list_count(sort="id") == 0

        # Currently, logged in as  e2e_user.
        # e2e_user will be the admin of the data source
        # no collaborators assigned to the data source but create as public
        self.create_random_datasource(e2e_client,
                                      name=class_context.generate_name(5),
                                      public=True,
                                      credentials=self.storage_credentials_user_rw)

        # list as temp_user
        temp_user_cnvrg = Cnvrg(domain=domain, email=e2e_temp_user["email"], password=e2e_temp_user["password"])
        datasources_count = temp_user_cnvrg.datasources.list_count(sort="id")
        assert datasources_count == 1
        ds = next(temp_user_cnvrg.datasources.list())
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        assert e2e_temp_user["email"] not in ds.collaborators

    # No Minio
    # def test_list_type(self, class_context, e2e_client):
    #     # Clean the db before testing list
    #     TestDatasources.cleanup(class_context.prefix)
    #     assert e2e_client.datasources.list_count(sort="id") == 0
    #
    #     # We create names with class prefix so that the cleanup will find them.
    #     # Create 2 AWS datasources
    #     name1 = class_context.generate_name(5)
    #     datasource = e2e_client.datasources.create(name=name1,
    #                                              storage_type=StorageTypes.S3,
    #                                              bucket_name=self.storage_bucket_name,
    #                                              region=self.storage_region,
    #                                              credentials=self.storage_credentials_user_rw)
    #     name2 = class_context.generate_name(5)
    #     datasource = e2e_client.datasources.create(name=name2,
    #                                              storage_type=StorageTypes.S3,
    #                                              bucket_name=self.storage_bucket_name,
    #                                              region=self.storage_region,
    #                                              credentials=self.storage_credentials_user_rw)
    #
    #     # Create 2 Minio datasources
    #     name3 = class_context.generate_name(5)
    #     datasource = e2e_client.datasources.create(name=name3,
    #                                              storage_type=StorageTypes.MINIO,
    #                                              bucket_name=self.storage_bucket_name,
    #                                              region=self.storage_region,
    #                                              credentials=self.storage_credentials_user_rw)
    #     name4 = class_context.generate_name(5)
    #     datasource = e2e_client.datasources.create(name=name4,
    #                                              storage_type=StorageTypes.MINIO,
    #                                              bucket_name=self.storage_bucket_name,
    #                                              region=self.storage_region,
    #                                              credentials=self.storage_credentials_user_rw)
    #     aws_count = 0
    #     minio_count = 0
    #     for ds in e2e_client.datasources.list(type=StorageTypes.S3):
    #         aws_count += 1
    #     for ds in e2e_client.datasources.list(type=StorageTypes.MINIO):
    #         minio_count += 1
    #     assert aws_count == 2
    #     assert minio_count == 2
    # endregion

    # region Add Collaborator
    # region test validity of collaborators
    def test_add_collaborator(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw)
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        ds.add_collaborator(e2e_temp_user['email'])
        ds.reload()
        assert len(ds.collaborators) == 2
        assert e2e_user['email'] in ds.collaborators
        assert e2e_temp_user['email'] in ds.collaborators

    def test_add_not_authorized_collaborator(self, class_context, e2e_client, e2e_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw)
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

        with pytest.raises(CnvrgHttpError) as exception_info:
            ds.add_collaborator('fake@mail.com')

        assert 'fake@mail.com' in str(exception_info.value)
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

    def test_add_existing_collaborator(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw)
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

        ds.add_collaborator(e2e_temp_user['email'])
        ds.reload()
        assert len(ds.collaborators) == 2
        assert e2e_user['email'] in ds.collaborators
        assert e2e_temp_user['email'] in ds.collaborators

        ds.add_collaborator(e2e_temp_user['email'])

        ds.reload()
        assert len(ds.collaborators) == 2
        assert e2e_user['email'] in ds.collaborators
        assert e2e_temp_user['email'] in ds.collaborators

    def test_add_admin_as_collaborator(self, class_context, e2e_client, e2e_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw)
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

        ds.add_collaborator(e2e_user['email'])

        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

    # endregion
    # region test authorizations
    def test_add_collaborator_as_non_admin(self, domain, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw)
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        # login as non admin user
        temp_user_cnvrg = Cnvrg(domain=domain, email=e2e_temp_user["email"], password=e2e_temp_user["password"])
        with pytest.raises(CnvrgHttpError) as exception_info:
            temp_ds = temp_user_cnvrg.datasources.get(ds.slug)
            temp_ds.add_collaborator(e2e_temp_user['email'])
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        assert 'Unauthorized' in str(exception_info.value)

    # endregion
    # endregion

    # region Remove Collaborator

    def test_remove_collaborator(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw
                                           )
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        ds.add_collaborator(e2e_temp_user['email'])
        ds.reload()
        assert len(ds.collaborators) == 2
        assert e2e_user['email'] in ds.collaborators
        assert e2e_temp_user['email'] in ds.collaborators
        ds.remove_collaborator(e2e_temp_user['email'])
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

    # region test validity of collaborators

    def test_remove_user_not_in_org(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw
                                           )
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        with pytest.raises(CnvrgHttpError) as exception_info:
            ds.remove_collaborator('fake@mail.com')
        assert 'fake@mail.com' in str(exception_info.value)
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

    def test_remove_user_not_assigned_to_the_datasource(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw
                                           )
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        with pytest.raises(CnvrgHttpError) as exception_info:
            ds.remove_collaborator(e2e_temp_user['email'])
        assert e2e_temp_user['email'] in str(exception_info.value)
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

    def test_remove_admin(self, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw)
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        with pytest.raises(CnvrgHttpError) as exception_info:
            ds.remove_collaborator(e2e_user['email'])
        assert e2e_user['email'] in str(exception_info.value)
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators

    # endregion

    # region test authorizations
    def test_remove_collaborator_as_non_admin(self, domain, class_context, e2e_client, e2e_user, e2e_temp_user):
        ds = self.create_random_datasource(e2e_client,
                                           name=class_context.generate_name(5),
                                           credentials=self.storage_credentials_user_rw
                                           )
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        # login as non admin user
        temp_user_cnvrg = Cnvrg(domain=domain, email=e2e_temp_user["email"], password=e2e_temp_user["password"])
        with pytest.raises(CnvrgHttpError) as exception_info:
            temp_ds = temp_user_cnvrg.datasources.get(ds.slug)
            temp_ds.remove_collaborator(e2e_temp_user['email'])
        ds.reload()
        assert len(ds.collaborators) == 1
        assert e2e_user['email'] in ds.collaborators
        assert 'Unauthorized' in str(exception_info.value)

    # endregion

    # endregion

    def test_list_count(self, class_context, e2e_client):
        # We create names with class prefix so that the cleanup will find them.
        count_before = e2e_client.datasources.list_count()
        number_of_obj_to_create = 23
        for i in range(number_of_obj_to_create):
            self.create_random_datasource(e2e_client,
                                          name=class_context.generate_name(5),
                                          credentials=self.storage_credentials_user_rw)

        count_after = e2e_client.datasources.list_count()
        assert count_after == count_before + number_of_obj_to_create

    # region HELPERS
    def create_random_datasource(self, client, name, collaborators=None, public=None, credentials=None):
        if collaborators:
            datasource = client.datasources.create(name=name,
                                                   storage_type=self.storage_type,
                                                   bucket_name=self.storage_bucket_name,
                                                   path=self.storage_path,
                                                   endpoint=self.storage_endpoint,
                                                   region=self.storage_region,
                                                   credentials=credentials,
                                                   collaborators=collaborators,
                                                   public=public)
        else:
            datasource = client.datasources.create(name=name,
                                                   storage_type=self.storage_type,
                                                   bucket_name=self.storage_bucket_name,
                                                   path=self.storage_path,
                                                   endpoint=self.storage_endpoint,
                                                   region=self.storage_region,
                                                   credentials=credentials,
                                                   public=public)
        return datasource
    # endregion



import os
import threading
import time
from queue import Empty, Full, Queue

from cnvrgv2.data.clients.storage_client_factory import storage_client_factory
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgRetryError


class DataLoader:
    def __init__(self, dataset, batch_size=1, num_workers=40, prefetch=3000, chunk_size=1000, override=False):
        """
        Python iterable over cnvrg datasets.
        Provides a file iterator that downloads files from dataset in real-time in order to parallelize data processing
        and dataset cloning. the data loader allows allows to batch files together and define number of files to
        prefetch
        @param dataset: Cnvrg dataset object
        @param batch_size: Iterator batch size - for values over 1 a list will be returned
        @param num_workers: Number of threads to download
        @param prefetch: Prefetch factor for files
        @param chunk_size: File meta chunk size to fetch from the server
        @param override: Override existing files
        """
        # Init the storage client
        self.dataset = dataset
        self.storage_client = storage_client_factory(refresh_function=dataset.storage_meta_refresh_function())

        # Handle batch input
        if not isinstance(batch_size, int) or batch_size < 1:
            raise CnvrgArgumentsError("Batch sise must be an integer greater than 0")
        self.batch_size = batch_size

        # Init file stats
        self.total_files = dataset.num_files
        self.files_processed = 0

        # Init file queues
        if not isinstance(prefetch, int) or prefetch < 1:
            raise CnvrgArgumentsError("Prefetch factor must be an integer greater than 0")
        if not isinstance(chunk_size, int) or chunk_size < 1:
            raise CnvrgArgumentsError("Chunk size must be an integer greater than 0")
        if not isinstance(num_workers, int) or num_workers < 1:
            raise CnvrgArgumentsError("Number of workers must be an integer greater than 0")

        self.override = override
        self.prefetch = prefetch
        self.chunk_size = chunk_size
        self.file_queue = Queue(self.prefetch)
        self.download_queue = Queue(self.prefetch)

        # Create a thread event in order to exit download when needed
        self.download_active = threading.Event()
        self.download_active.set()

        # Create a thread-safe lock
        self.download_lock = threading.Lock()

        # Create collector thread which fetches file chunks from the server
        self.file_index = 1
        self.collector_thread = threading.Thread(target=self.file_collector)
        self.collector_thread.start()

        # Create downloader threads to parallelize s3 file download
        self.downloader_threads = []
        for i in range(num_workers):
            t = threading.Thread(target=self.file_downloader)
            t.start()
            self.downloader_threads.append(t)

    def __iter__(self):
        return self

    def __next__(self):
        """
        Returns a single file meta object if batch_size is 1, otherwise returns a list of file meta objects
        @return: dict/list
        """
        file_batch = []

        # Attempt to batch files together
        while (len(file_batch) < self.batch_size) and (self.files_processed < self.total_files):
            file = self.file_queue.get()

            with self.download_lock:
                self.files_processed += 1

            self.file_queue.task_done()
            file_batch.append(file)

        if len(file_batch) > 0 and self.batch_size > 1:
            # We return list if the batch size is greater than 1
            return file_batch
        elif len(file_batch) > 0 and self.batch_size == 1:
            # We return a single object if the batch size is equal to 1
            return file_batch[0]
        else:
            # If the file batch is empty it means we processed all of the files and should clear download
            self.clear()
            raise StopIteration

    def __enter__(self):
        """
        Support context manager
        @return: self
        """
        return self

    def __exit__(self, *exc):
        """
        Ensure data loader threads will be cleaned in case of an error
        @param exc: The error that triggered exit
        @return:
        """
        # TODO: Handle exception?
        self.clear()

    def clear(self):
        """
        Clear the threads used to download files
        @return: none
        """
        # Clear download threads
        self.download_active.clear()
        self.collector_thread.join()

        for t in self.downloader_threads:
            t.join()

    def file_collector(self):
        """
        The function that handles collecting files-to-download metadata from the server
        @return: None
        """
        while self.file_index < self.total_files and self.download_active.is_set():
            end_index = self.file_index + self.chunk_size

            # Attempt to retrieve file chunk from the server
            try:
                files_to_download = self.dataset.get_files_by_range(
                    start=self.file_index,
                    end=end_index
                )
            except Exception:
                files_to_download = []
                print("Could not download files {}-{}".format(self.file_index, end_index))
                # TODO: Change exception handling? add retry?

                # If we could not get the chunk we still count it as processed
                with self.download_lock:
                    self.files_processed += self.chunk_size
            finally:
                self.file_index = end_index

            # Attempt to put the new files in the download queue, non-blocking in case we want to break iterator loop
            for file in files_to_download:
                while self.download_active.is_set():
                    try:
                        self.download_queue.put_nowait(file)
                        break
                    except Full:
                        time.sleep(0.5)

    def file_downloader(self):
        """
        Handles downloading files from the remote object storage.
        @return: None
        """
        # Run as long as we have files to download
        while self.download_active.is_set():
            try:
                # Get file non-blocking way, otherwise thread will hang forever
                file = self.download_queue.get_nowait()
                local_path = "{}/{}".format(self.dataset.working_dir, file["fullpath"])
                if self.override or not os.path.exists(local_path):
                    self.storage_client.download_single_file(local_path, file["object_path"])
                file["local_path"] = local_path
                self.download_queue.task_done()

                self.file_queue.put(file)
            except Empty:
                time.sleep(0.5)
            except CnvrgRetryError:
                # If we could not download the file we still count it as processed
                with self.download_lock:
                    self.files_processed += 1
            except Exception as e:
                # Should not be possible to enter here, safeguard against deadlock
                print("An unhandled exception in downloader thread has occurred:")
                print(e)
                with self.download_lock:
                    self.files_processed += 1



import pytest

from cnvrgv2.errors import CnvrgHttpError


class TestCompute:

    def test_get_create_pytorch_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        new_kube_template = templates.create("pytourch_kube", 2, 2, type="pytorch", worker_num_executors=1,
                                             gpu=2.0, worker_labels="test_1=2,test_2=mock",
                                             worker_taints="test_1=2,cnvrg=kube")

        kube_template = templates.get(new_kube_template.slug)
        assert kube_template.title == "pytourch_kube"
        assert kube_template.cpu == 2
        assert kube_template.memory == 2

    def test_get_create_open_mpi_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        new_kube_template = templates.create("mpi_kube", cpu=2, memory=2, gpu=1,
                                             labels="test_33=2,test_2=22", type="mpi", worker_num_executors=1,
                                             worker_cpu=2.0, worker_memory=2.0, worker_gpu=1, shm=2,
                                             worker_labels="test_1=2,test_2=mock", worker_taints="test_1=2,cnvrg=kube")
        kube_template = templates.get(new_kube_template.slug)
        assert kube_template.title == "mpi_kube"
        assert dict(kube_template.worker)['data']['attributes']['cpu'] == 2
        assert dict(kube_template.worker)['data']['attributes']['labels'] == 'test_1: 2, test_2: mock'
        assert dict(kube_template.worker)['data']['attributes']['gpu'] == 1
        assert dict(kube_template.worker)['data']['attributes']['memory'] == 2
        assert dict(kube_template.worker)['data']['attributes']['num_executors'] == 1
        assert kube_template.labels == "test_33: 2, test_2: 22"
        assert kube_template.cpu == 2
        assert kube_template.memory == 2
        assert kube_template.gpu == 1

    def test_get_create_ray_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        new_kube_template = templates.create("ray_kube", cpu=2, memory=2, gpu=1,
                                             labels="test_33=2,test_2=22", type="ray", worker_num_executors=1,
                                             worker_cpu=2.0, worker_memory=2.0, worker_gpu=1, shm=2,
                                             worker_labels="test_1=2,test_2=mock", worker_taints="test_1=2,cnvrg=kube")
        kube_template = templates.get(new_kube_template.slug)

        assert kube_template.title == "ray_kube"
        assert dict(kube_template.worker)['data']['attributes']['cpu'] == 2
        assert dict(kube_template.worker)['data']['attributes']['labels'] == 'test_1: 2, test_2: mock'
        assert dict(kube_template.worker)['data']['attributes']['gpu'] == 1
        assert dict(kube_template.worker)['data']['attributes']['memory'] == 2
        assert dict(kube_template.worker)['data']['attributes']['num_executors'] == 1
        assert kube_template.labels == "test_33: 2, test_2: 22"
        assert kube_template.cpu == 2
        assert kube_template.memory == 2
        assert kube_template.gpu == 1

    def test_get_create_ray_modin_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        new_kube_template = templates.create("ray_modin_kube", cpu=2, memory=2, gpu=1,
                                             labels="test_33=2,test_2=22", type="ray_modin", worker_num_executors=1,
                                             worker_cpu=2.0, worker_memory=2.0, worker_gpu=1, shm=2,
                                             worker_labels="test_1=2,test_2=mock", worker_taints="test_1=2,cnvrg=kube")
        kube_template = templates.get(new_kube_template.slug)

        assert kube_template.title == "ray_modin_kube"
        assert dict(kube_template.worker)['data']['attributes']['cpu'] == 2
        assert dict(kube_template.worker)['data']['attributes']['labels'] == 'test_1: 2, test_2: mock'
        assert dict(kube_template.worker)['data']['attributes']['gpu'] == 1
        assert dict(kube_template.worker)['data']['attributes']['memory'] == 2
        assert dict(kube_template.worker)['data']['attributes']['num_executors'] == 1
        assert kube_template.labels == "test_33: 2, test_2: 22"
        assert kube_template.cpu == 2
        assert kube_template.memory == 2
        assert kube_template.gpu == 1

    def test_get_create_spark_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        spark_config = "spark.executor.memory=2\nspark.executor.cores=4\nspark.executor.instances=2"
        new_kube_template = templates.create("spark_kube", cpu=2, memory=2, gpu=1,
                                             labels="test_33=2,test_2=22", gaudi=2, huge_pages=500,
                                             type="spark_on_kubernetes",
                                             spark_config=spark_config)

        kube_template = templates.get(new_kube_template.slug)

        assert kube_template.title == "spark_kube"
        assert kube_template.labels == "test_33: 2, test_2: 22"
        assert kube_template.cpu == 2
        assert kube_template.memory == 2
        assert kube_template.gpu == 1
        assert kube_template.gaudi == 2
        assert kube_template.spark_config == spark_config

    def test_list_templates(self, e2e_cluster):
        templates_client = e2e_cluster.templates
        spark_config = "spark.executor.memory=2\nspark.executor.cores=4\nspark.executor.instances=2"
        spark_kube_template = templates_client.create("spark_kube", cpu=2, memory=2, gpu=1,
                                                      labels="test_33=2,test_2=22", gaudi=2, huge_pages=500,
                                                      type="spark_on_kubernetes",
                                                      spark_config=spark_config)

        ray_modin_kube_template = templates_client.create("ray_modin_kube", cpu=2, memory=2, gpu=1,
                                                          labels="test_33=2,test_2=22", type="ray_modin",
                                                          worker_num_executors=1,
                                                          worker_cpu=2.0, worker_memory=2.0, worker_gpu=1, shm=2,
                                                          worker_labels="test_1=2,test_2=mock",
                                                          worker_taints="test_1=2,cnvrg=kube")

        all_templates = [template.title for template in templates_client.list()]

        assert spark_kube_template.title in all_templates
        assert ray_modin_kube_template.title in all_templates

    def test_update_spark_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        spark_config = "spark.executor.memory=2\nspark.executor.cores=4\nspark.executor.instances=2"
        new_spark_template = templates.create("spark_kube", cpu=2, memory=2, gpu=1,
                                              labels="test_33=2,test_2=22", gaudi=2, huge_pages=500,
                                              type="spark_on_kubernetes",
                                              spark_config=spark_config)

        kube_template = templates.get(new_spark_template.slug)
        assert kube_template.title == "spark_kube"
        assert kube_template.labels == "test_33: 2, test_2: 22"
        assert kube_template.cpu == 2
        assert kube_template.memory == 2
        assert kube_template.gpu == 1
        assert kube_template.gaudi == 2
        assert kube_template.spark_config == spark_config

        updated_spark_config = "spark.executor.memory=3\nspark.executor.cores=5\nspark.executor.instances=6"
        kube_template.update(title="spark_kube2", cpu=4, memory=4, gpu=2, labels="test_33=4,test_2=44", gaudi=4,
                             huge_pages=1000,
                             spark_config=updated_spark_config)

        assert kube_template.title == "spark_kube2"
        assert kube_template.labels == "test_33: 4, test_2: 44"
        assert kube_template.cpu == 4
        assert kube_template.memory == 4
        assert kube_template.gpu == 2
        assert kube_template.gaudi == 4
        assert kube_template.spark_config == updated_spark_config

    def test_update_empty_gpu_spark_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        spark_config = "spark.executor.memory=2\nspark.executor.cores=4\nspark.executor.instances=2"
        new_spark_template = templates.create("spark_kube", cpu=2, memory=2, gpu=1,
                                              labels="test_33=2,test_2=22", gaudi=2, huge_pages=500,
                                              type="spark_on_kubernetes",
                                              spark_config=spark_config)

        kube_template = templates.get(new_spark_template.slug)
        assert kube_template.title == "spark_kube"
        assert kube_template.labels == "test_33: 2, test_2: 22"
        assert kube_template.cpu == 2
        assert kube_template.memory == 2
        assert kube_template.gpu == 1
        assert kube_template.gaudi == 2
        assert kube_template.spark_config == spark_config

        kube_template.update(title="spark_kube2", cpu=4, memory=4, gpu=0, labels="", gaudi=4,
                             huge_pages=1000,
                             spark_config="")
        assert kube_template.title == "spark_kube2"
        assert kube_template.labels is None
        assert kube_template.cpu == 4
        assert kube_template.memory == 4
        assert kube_template.gpu == 0
        assert kube_template.gaudi == 4
        assert kube_template.spark_config == ""

    def test_delete_spark_template(self, e2e_cluster):
        templates = e2e_cluster.templates
        spark_config = "spark.executor.memory=2\nspark.executor.cores=4\nspark.executor.instances=2"
        new_spark_template = templates.create("spark_kube", cpu=2, memory=2, gpu=1,
                                              labels="test_33=2,test_2=22", gaudi=2, huge_pages=500,
                                              type="spark_on_kubernetes",
                                              spark_config=spark_config)

        kube_template = templates.get(new_spark_template.slug)
        kube_template.delete()
        with pytest.raises(CnvrgHttpError) as exception_info:
            kube_template.title
        assert "Could not find compute" in str(exception_info.value)



from cnvrgv2.modules.datasource import StorageTypes
import pytest
from cnvrgv2.config import GLOBAL_CNVRG_PATH
from pathlib import Path
import os
from dotenv import load_dotenv
from tests.long_running_tests.datasources.conftest import datasource_test_bukcet_name, datasource_test_region


class TestDatasourcesFilesOperations:
    storage_type = StorageTypes.S3
    storage_bucket_name = None
    storage_endpoint = None
    storage_region = None
    storage_credentials_user_rw = None
    storage_credentials_user_ro = None
    datasource_rw = None
    datasource_ro = None
    datasource_rw_no_path = None
    config_env_vars = ['DATASOURCE_TEST_USER_RW_ACCESS_KEY_ID', 'DATASOURCE_TEST_USER_RW_SECRET_ACCESS_KEY',
                       'DATASOURCE_TEST_USER_RO_ACCESS_KEY_ID', 'DATASOURCE_TEST_USER_RO_SECRET_ACCESS_KEY']

    @classmethod
    def set_datasource_aws_config(cls):
        # Path to the AWS config file
        config_file = os.path.join(GLOBAL_CNVRG_PATH, '.env')

        # load from config file if running locally.
        # Env vars should exist when running through github actions
        if Path(config_file).is_file():
            load_dotenv(config_file)

        bucket_name = datasource_test_bukcet_name
        region = datasource_test_region
        user1_access_key_id = os.getenv(key='DATASOURCE_TEST_USER_RW_ACCESS_KEY_ID')
        user1_secret_access_key = os.getenv(key='DATASOURCE_TEST_USER_RW_SECRET_ACCESS_KEY')
        user2_access_key_id = os.getenv(key='DATASOURCE_TEST_USER_RO_ACCESS_KEY_ID')
        user2_secret_access_key = os.getenv(key='DATASOURCE_TEST_USER_RO_SECRET_ACCESS_KEY')

        cls.storage_bucket_name = bucket_name
        cls.storage_region = region
        cls.storage_credentials_user_rw = {
            "access_key_id": user1_access_key_id,
            "secret_access_key": user1_secret_access_key
        }
        cls.storage_credentials_user_ro = {
            "access_key_id": user2_access_key_id,
            "secret_access_key": user2_secret_access_key
        }

    @classmethod
    def create_datasources(cls, class_context, e2e_client):
        path = class_context.generate_name(5)
        cls.datasource_rw = e2e_client.datasources.create(name=class_context.generate_name(5),
                                                          storage_type=cls.storage_type,
                                                          path=path,
                                                          bucket_name=cls.storage_bucket_name,
                                                          region=cls.storage_region,
                                                          credentials=cls.storage_credentials_user_rw)
        cls.datasource_rw_no_path = e2e_client.datasources.create(name=class_context.generate_name(5),
                                                                  storage_type=cls.storage_type,
                                                                  bucket_name=cls.storage_bucket_name,
                                                                  region=cls.storage_region,
                                                                  credentials=cls.storage_credentials_user_rw)

        cls.datasource_ro = e2e_client.datasources.create(name=class_context.generate_name(5),
                                                          storage_type=cls.storage_type,
                                                          path=path,
                                                          bucket_name=cls.storage_bucket_name,
                                                          region=cls.storage_region,
                                                          credentials=cls.storage_credentials_user_ro)

    @classmethod
    def clear_env(cls):
        # Clear env varaiables
        for variable_name in cls.config_env_vars:
            if variable_name in os.environ:
                os.environ.pop(variable_name)

    @classmethod
    def clear_bucket_path(cls):
        cls.remove_files(datasource=cls.datasource_rw)

    @classmethod
    @pytest.fixture(scope='class', autouse=True)
    def setup_and_teardown_class(cls, class_context, e2e_client):
        # Setup Class - Executed once per test class
        cls.set_datasource_aws_config()
        # Skip tests if config is missing
        if any(var is None for var in [cls.storage_bucket_name, cls.storage_region,
                                       cls.storage_credentials_user_rw['access_key_id'],
                                       cls.storage_credentials_user_rw['secret_access_key'],
                                       cls.storage_credentials_user_ro['access_key_id'],
                                       cls.storage_credentials_user_ro['secret_access_key']]):
            pytest.skip("Skipping tests due to CONFIGURATION MISSING")
        cls.create_datasources(class_context, e2e_client)
        yield
        # Teardown Class - Executed once after all test methods in the class
        cls.clear_env()
        # Every test should clear up the path,
        # but we also clear path in teardown in case some tests fail before they clear up.
        cls.clear_bucket_path()

    # region Helpers
    def count_paginator(self, paginator):
        file_count = 0
        for page in paginator:
            for file_name in page:
                file_count += 1
        return file_count

    @classmethod
    def remove_files(self, datasource):
        paginator = datasource.list_objects()
        for page in paginator:
            for file_name in page:
                datasource.remove_file(file_name)
    # endregion



import os
import traceback
from functools import wraps


def suppress_exceptions(f):
    """
    Used to suppress exceptions across the sdk to not interrupt client script runs
    @param f: the function we wish to suppress
    @return: suppressed function
    """

    suppress = os.environ.get('CNVRG_SUPPRESS_EXCEPTIONS')

    @wraps(f)
    def func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            if suppress:
                traceback.print_exc()
            else:
                raise e
    return func



import pytest

from cnvrgv2.errors import CnvrgHttpError


class TestResources:
    def test_create_cluster_success(self, random_name, e2e_client):
        cluster_name = random_name(5)
        attributes = {
            "domain": "test-domain.cicd.cnvrg.me",
            "scheduler": "cnvrg_scheduler",
            "namespace": "test-namespace"
        }
        cluster = e2e_client.clusters.create(
            kube_config_yaml_path="assets/kube_config.yaml",
            resource_name=cluster_name,
            domain=attributes["domain"],
            scheduler=attributes["scheduler"],
            namespace=attributes["namespace"],
            https_scheme=True,
            persistent_volumes=True,
            gaudi_enabled=True
        )

        assert cluster.title == cluster_name
        assert cluster.domain == attributes["domain"]
        assert cluster.scheduler == attributes["scheduler"]
        assert cluster.namespace == attributes["namespace"]
        assert cluster.https_scheme is True
        assert cluster.persistent_volumes is True

    def test_update_cluster_success(self, random_name, e2e_client, e2e_cluster):
        cluster_name = random_name(5)
        attributes = {"domain": "test-domain.cicd.cnvrg.me",
                      "scheduler": "volcano",
                      "namespace": "test-namespace-new"}
        e2e_cluster.update(domain=attributes["domain"])
        e2e_cluster.reload()
        assert e2e_cluster.domain == attributes["domain"]

        e2e_cluster.update(resource_name=cluster_name, scheduler=attributes["scheduler"],
                           namespace=attributes["namespace"])
        e2e_cluster.reload()
        assert e2e_cluster.title == cluster_name
        assert e2e_cluster.scheduler == attributes["scheduler"]
        assert e2e_cluster.namespace == attributes["namespace"]

        e2e_cluster.update(https_scheme=True, persistent_volumes=True)
        e2e_cluster.reload()
        assert e2e_cluster.https_scheme is True
        assert e2e_cluster.persistent_volumes is True

    def test_delete_cluster_success(self, random_name, e2e_client, e2e_cluster):
        slug = e2e_cluster.slug
        e2e_cluster.delete()
        cluster = e2e_client.clusters.get(slug)
        with pytest.raises(CnvrgHttpError) as exception_info:
            cluster.title
        assert "The requested resource could not be found" in str(exception_info.value)



from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.workflows.endpoint.endpoint import Endpoint
from cnvrgv2.proxy import Proxy
from cnvrgv2.utils.json_api_format import JAF


class TestEndpoint:
    def test_stopped_endpoint_update_replicas_error(self, empty_context):
        pass
        # e = Endpoint(context=empty_context, slug="end-slug", attributes={'status': 'stopped'})
        # with pytest.raises(CnvrgError) as exception_info:
        #     e.update_replicas(2, 2)
        # assert 'not running' in str(exception_info.value)

    def test_stopped_endpoint_restart_attribute(self, empty_context):
        try:
            e = Endpoint(context=empty_context, slug="end-slug", attributes={'status': 'stopped'})
            e.restart()
        except AttributeError as err:
            assert "'Endpoint' object has no attribute 'restart'" in str(err)

    def test_send_logs_wrong_type(self, empty_context):
        try:
            e = Endpoint(context=empty_context, slug="end-slug", attributes={'status': 'stopped'})
            e.log(logs="test", log_level="wrong_type")
        except CnvrgArgumentsError as err:
            assert "Wrong attribute type, expected logs level enum, got wrong_type" in str(err)

    def test_send_logs_wrong_message_type(self, empty_context):
        try:
            e = Endpoint(context=empty_context, slug="end-slug", attributes={'status': 'stopped'})
            e.log(logs=1)
        except CnvrgArgumentsError as err:
            assert "Wrong attribute type, expected str, got <class 'int'>" in str(err)

    def test_endpoint_get_predictions(self, mocker, empty_context):
        def mock_response(*args, **kwargs):
            data = {'predictions': [{'model': '1',
                                     'start_time': '2023-01-02 10:34:59',
                                     'input': 'your_input_params',
                                     'output': 'your_input_params',
                                     'elapsed_time': 3,
                                     'offset': 1672655699815},
                                    {'model': '1',
                                     'start_time': '2023-01-02 10:35:59',
                                     'input': 'your_input_params1',
                                     'output': 'your_input_params1',
                                     'elapsed_time': 3,
                                     'offset': 1672655699511},
                                    ], 'offset': 1672655698141, 'total': 2}

            return JAF(response={"data": {"attributes": data}})

        mocker.patch.object(Proxy, 'call_api', side_effect=mock_response)
        e = Endpoint(context=empty_context, slug="end-slug")
        res = e.get_predictions()

        assert res["total"] == 2



from cnvrgv2.config import routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.labels.utils import validate_labels_params
from cnvrgv2.utils.url_utils import urljoin


class Label(DynamicAttributes):
    available_attributes = {
        "id": str,
        "name": str,
        "kind": str,
        "color_name": str,
    }

    def __init__(self, context=None, name=None, kind=None, attributes=None):
        self._context = Context(context=context)

        if name and kind:
            self._context.set_scope(SCOPE.LABEL, {"name": name, "kind": kind})

        scope = self._context.get_scope(SCOPE.LABEL)

        self._proxy = Proxy(context=self._context)
        self._route = routes.LABELS_BASE.format(scope["organization"])
        self._attributes = attributes or {}

    def delete(self):
        """
        Deletes the current label
        @return: None
        """
        url = urljoin(self._route, self.id)
        self._proxy.call_api(route=url, http_method=HTTP.DELETE)

    def save(self):
        return self.update(name=self.name, color_name=self.color_name)

    def update(self, name=None, color_name=None):
        if name:
            self.name = name

        if color_name:
            self.color_name = color_name

        validate_labels_params(self.name, self.kind, self.color_name)
        attributes = {
            "id": self.id,
            "name": self.name,
            "kind": self.kind,
            "color_name": self.color_name,
        }

        self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type="label", attributes=attributes)
        )
        return self



from cnvrgv2.context import Context


class MockResponse:
    def __init__(self, status, json):
        self.status_code = status
        self._json = json

    def json(self):
        return self._json


class TestCnvrg:
    def test_set_organization(self, mocker, empty_cnvrg):
        new_org = "new-org"
        mocker.patch.object(Context, "set_scope")

        assert empty_cnvrg._organization != new_org
        empty_cnvrg.set_organization(new_org)
        assert empty_cnvrg._organization == new_org



import os
import random
import shutil
import string
import argparse
import requests
from cnvrgv2 import Cnvrg, EndpointKind, EndpointEnvSetup, WebappType, errors
from cnvrgv2.errors import CnvrgHttpError
from cnvrgv2.modules.users import ROLES
from cnvrgv2.modules.users.users_client import UsersClient
from cnvrgv2.modules.workflows.workflow_utils import WorkflowUtils, WorkflowStatuses


def random_string(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))


class LoadCluster:
    LOAD_CLUSTER_FILES_DIR = './.LOAD_CLUSTER_FILES_DIR'
    DEFAULT_NUMBERS = 2000
    DEFAULT_FILES = 1000000  # 1 million
    DEFAULT_USERS = 20
    HELP_DESCRIPTION = """
    This script is used to create a cluster with many workflows, datasets, projects, files etc.
    It is useful for load testing and can be used programmatically
    too just by importing the script and choosing the list of commands.
    """

    def __init__(self,
                 domain=None,
                 email=None,
                 password=None,
                 stop=True,
                 num_workflows=DEFAULT_NUMBERS,
                 num_datasets=DEFAULT_NUMBERS,
                 num_projects=DEFAULT_NUMBERS,
                 num_files=DEFAULT_FILES,
                 num_endpoint_charts=DEFAULT_NUMBERS,
                 num_users=DEFAULT_USERS
                 ):
        parser = argparse.ArgumentParser(description=self.HELP_DESCRIPTION)
        parser.add_argument("-d", "--domain", help="choose a domain to run on", type=str, default=domain)
        parser.add_argument("-e", "--email", help="email to login into that domain", type=str, default=email)
        parser.add_argument("-p", "--password", help="password to login into that domain", type=str,
                            default=password)
        parser.add_argument("-nw", "--num-workflows", help="number workflows to create (defaults to 2000)",
                            default=num_workflows, type=int)
        parser.add_argument("-s", "--stop", help="stop workflows after creation (defaults to True)",
                            default=stop, type=bool)
        parser.add_argument("-nd", "--num-datasets", help="number of datasets to create (defaults to 2000)",
                            default=num_datasets, type=int)
        parser.add_argument("-np", "--num-projects", help="number of projects to create (defaults to 2000)",
                            default=num_projects, type=int)
        parser.add_argument("-nf", "--num-files", help="number of files to create into a dataset (defaults to 2000)",
                            default=num_files, type=int)
        parser.add_argument("-nep", "--num-endpoint-charts",
                            help="number of end point chart to create (defaults to 2000)",
                            default=num_endpoint_charts, type=int)
        parser.add_argument("-nu", "--num-users",
                            help="number of users to create for each role (defaults to 20 for each role)",
                            default=num_users, type=int)

        self.args = parser.parse_args()

        self.cnvrg = Cnvrg(domain=self.args.domain, email=self.args.email, password=self.args.password)
        self.proj = self.cnvrg.projects.get('load_proj')
        try:
            # if project doesn't exist it will throw an exception
            self.proj.title
        except CnvrgHttpError:
            self.proj = self.cnvrg.projects.create('load_proj')

        # Create a dir for storing generated files
        if os.path.exists(self.LOAD_CLUSTER_FILES_DIR):
            if os.path.isdir(self.LOAD_CLUSTER_FILES_DIR):
                shutil.rmtree(self.LOAD_CLUSTER_FILES_DIR)
            else:
                os.remove(self.LOAD_CLUSTER_FILES_DIR)

        os.makedirs(self.LOAD_CLUSTER_FILES_DIR)

    def create_cluster(self):
        self._try_to_run("create_many_datasets")
        self._try_to_run("create_many_projects")
        self._try_to_run("create_large_load_dataset")
        self._try_to_run("create_large_project")
        self._try_to_run("create_experiment_with_many_charts")
        self._try_to_run("create_experiment_with_many_artifacts")
        self._try_to_run("create_endpoint_charts")
        self._try_to_run("create_many_users")
        self.destroy()

    def create_many_users(self):
        print("Creating users..")
        random_suffix = random_string(5)
        roles = [ROLES.ADMIN, ROLES.MEMBER, ROLES.MANAGER, ROLES.REVIEWER]

        for role in roles:
            for i in range(self.args.num_users):
                username = str(role) + random_suffix + str(i)
                email = username + "@cnvrg.io"
                password = "qwe123"
                uc = UsersClient(domain=self.args.domain)
                uc.register(username=username, email=email, password=password)
                self.cnvrg.members.add(email, role)
        print("Done creating users.")

    def create_large_project(self):
        print("Creating large project..")
        arguments = {
            "templates": ["small"],
            "command": "python3 -c 1+1"
        }

        for i in range(self.args.num_workflows):
            self._progress_print(i + 1, self.args.num_workflows)
            job = self.proj.workspaces.create(**arguments)
            if self.args.stop:
                job.stop()
            job = self.proj.experiments.create(**{**arguments, "sync_before": False, "sync_after": False})
            if self.args.stop:
                job.stop()
            job = self.proj.endpoints.create(
                title=random_string(10) + str(i),
                templates=["small"],
                kind=EndpointKind.WEB_SERVICE,
                file_name="non-existing.py",
                function_name="predict",
                env_setup=EndpointEnvSetup.PYTHON3
            )
            if self.args.stop:
                job.stop()
            job = self.proj.webapps.create(
                title=random_string(10) + str(i),
                templates=["small"],
                file_name="non-existing.py",
                webapp_type=WebappType.SHINY
            )
            if self.args.stop:
                job.stop()

        print("Done creating large project.")

    def create_many_projects(self):
        print("Creating many projects..")
        for i in range(self.args.num_projects):
            self._progress_print(i + 1, self.args.num_projects)
            self.cnvrg.projects.create("load-project-" + str(i))
        print("Done creating many projects.")

    def create_many_datasets(self):
        print("Creating datasets..")
        for i in range(self.args.num_datasets):
            self._progress_print(i + 1, self.args.num_datasets)
            self.cnvrg.datasets.create("load-datasets-" + str(i))
        print("Done creating datasets.")

    def create_large_load_dataset(self):
        print("Creating large load dataset..")
        dataset = self.cnvrg.datasets.create("large_load_dataset")
        for i in range(self.args.num_files):
            self._progress_print(i + 1, self.args.num_files)
            file_path = "{}/test_file_{}".format(self.LOAD_CLUSTER_FILES_DIR, i)
            tags_file_path = file_path + "_tags.yml"
            with open(file_path, 'w+') as file:
                file.write(file_path)
            with open(tags_file_path, 'w+') as yml:
                yml.write("---\nid: \"{}\"\nsource: \"mega cluster script\"".format(i))

            dataset.put_files([file_path, tags_file_path])
        print("Done creating large load datasets.")

    def create_endpoint_charts(self):
        print("Creating endpoint charts..")
        self.proj.put_files(['examples/load_snippets/endpoint_charts.py'])
        endpoint = self.proj.endpoints.create(
            title=random_string(10),
            templates=["small"],
            kind=EndpointKind.WEB_SERVICE,
            file_name="examples/load_snippets/endpoint_charts.py",
            function_name="predict",
            env_setup=EndpointEnvSetup.PYTHON3
        )

        try:
            WorkflowUtils.wait_for_statuses(endpoint, [WorkflowStatuses.ONGOING])
        except errors.CnvrgFinalStateReached:
            return

        endpoint = self.proj.endpoints.get(endpoint.slug)

        for i in range(2000):
            self._progress_print(i + 1, 2000)
            requests.post(
                endpoint.endpoint_url,
                headers={
                    "Cnvrg-Api-Key": endpoint.api_key,
                },
                data={"doesn't": "even matter"}
            )
        print("Done creating endpoint charts.")

    def create_experiment_with_many_charts(self):
        print("Creating experiment with many charts..")
        self.proj.put_files(['examples/load_snippets/experiment_with_charts.py'])
        self.proj.experiments.create(
            title="all possible charts experiment",
            templates=["small"],
            command="python3 examples/load_snippets/experiment_with_charts.py",
            sync_before=False,
            sync_after=False
        )
        print("Done creating experiment with many charts.")

    def create_experiment_with_many_artifacts(self):
        print("Creating experiment with many artifacts..")
        self.proj.put_files(["examples/load_snippets/experiment_with_many_artifacts.py"])
        self.proj.experiments.create(
            title="a lot of artifacts experiment",
            templates=["small"],
            command="python3 examples/load_snippets/experiment_with_many_artifacts.py",
            sync_before=False,
            sync_after=False
        )
        print("Done creating experiment with many artifacts.")

    def destroy(self):
        shutil.rmtree(self.LOAD_CLUSTER_FILES_DIR)

    def _try_to_run(self, method_name: str):
        try:
            getattr(self, method_name)()

        except Exception as e:
            print("{} has failed, exception: {}".format(method_name, str(e)))

    def _progress_print(self, current, total):
        percentage = current / total
        progress = "<"
        remaining = int(percentage * 10)
        for i in range(remaining):
            progress += "="
        for i in range(10 - remaining):
            progress += "-"
        progress += "> {}/{}".format(current, total)

        print(progress, end="\r")

        # last progress report add new line
        if total == current:
            print()


def createLoadCuster():
    LoadCluster().create_cluster()


if __name__ == "__main__":
    cluster = LoadCluster()
    cluster.create_cluster()



from cnvrgv2.cnvrg import Cnvrg
from cnvrgv2.errors import ExceptionsConfig
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.modules.project import Project
from cnvrgv2.modules.workflows import Endpoint, EndpointEnvSetup, EndpointKind, Experiment, Webapp, WebappType, \
    Workspace
from cnvrgv2.utils.chart_utils import BarChart, Heatmap, LineChart, ScatterPlot



import os
import pytest
from cnvrgv2 import Project
from cnvrgv2.cnvrg import Cnvrg
from cnvrgv2.context import Context


@pytest.fixture(scope="function")
def empty_context(mocker):
    # Test that the load_credentials function was called on empty input
    def mock_context_init(obj, *args, **kwargs):
        # Set the credentials variables:
        obj.token = "empty"
        obj.domain = "empty"
        obj.user = "empty"

        # Set context variables:
        obj.project = "empty"
        obj.experiment = "empty"
        obj.organization = "empty"
        obj.dataset = "empty"
        obj.datasource = "empty"
        obj.flow = "empty"
        obj.resource = "empty"
        obj.sso_version = None

    mocker.patch.object(Context, "__init__", mock_context_init)

    return Context()


@pytest.fixture(scope="function")
def empty_cnvrg(mocker, empty_context):
    def mock_cnvrg_init(obj, *args, **kwargs):
        obj._domain = "domain"
        obj._email = "test@email.com"
        obj._password = "password"
        obj._organization = "example-org"

    mocker.patch.object(Cnvrg, "__init__", mock_cnvrg_init)
    cnvrg = Cnvrg()
    cnvrg._context = empty_context

    return cnvrg


@pytest.fixture(scope="function")
def sample_netrc(request, mocker):
    fake_netrc_home = os.path.expanduser("~")

    def clear_netrc():
        os.remove(fake_netrc_home + "/.netrc")

    mocker.patch.dict(os.environ, {"HOME": fake_netrc_home}, clear=True)
    netrc_content = [
        "machine cnvrg.io",
        "  login test@test.com",
        "  password sample_token"
    ]
    with open(fake_netrc_home + "/.netrc", "w+") as f:
        f.writelines('\n'.join(netrc_content))

    os.chmod(fake_netrc_home + "/.netrc", 0o600)
    request.addfinalizer(clear_netrc)


@pytest.fixture(scope="function")
def empty_project(empty_context):
    return Project(context=empty_context)



import click

from cnvrgv2 import EndpointKind, EndpointEnvSetup
from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.utils.log_utils import LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR, LOGS_TYPE_INFO, LOGS_TYPE_WARNING
from cnvrgv2.cli.utils.helpers import pretty_print_predictions


@click.group(name='endpoint')
def endpoint_group():
    pass


@endpoint_group.command()
@click.option('-t', '--title', prompt=messages.ENDPOINT_PROMPT_TITLE_NAME, help=messages.ENDPOINT_HELP_TITLE,
              required=True)
@click.option('-f', '--file_name', prompt=messages.ENDPOINT_PROMPT_FILE_NAME, help=messages.ENDPOINT_HELP_FILE_NAME,
              default=None)
@click.option('-fn', '--function_name', prompt=messages.ENDPOINT_PROMPT_FUNCTION_NAME,
              help=messages.ENDPOINT_HELP_FUNCTION_NAME,
              default=None)
@click.option('-k', '--kind', default=EndpointKind.WEB_SERVICE, help=messages.ENDPOINT_HELP_KIND)
@click.option('-e', '--env_setup', default=EndpointEnvSetup.PYTHON3, help=messages.ENDPOINT_HELP_ENV_SETUP)
@click.option('-te', '--templates', default=None, help=messages.ENDPOINT_HELP_TEMPLATES)
@click.option('-kb', '--kafka_brokers', default=None, help=messages.ENDPOINT_HELP_KAFKA_BROKERS)
@click.option('-kit', '--kafka_input_topics', default=None,
              help=messages.ENDPOINT_HELP_KAFKA_INPUT_TOPICS)
@click.option('-a', '--args', multiple=True, default=None, help=messages.ENDPOINT_HELP_ARGS)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def create(project,
           logger,
           title,
           file_name,
           function_name,
           kind, env_setup,
           templates,
           kafka_brokers,
           kafka_input_topics,
           args,
           project_slug):
    """
        create an endpoint
    """
    param_args = {}
    if args:
        param_args = dict([p.split('=') for p in args])

    kwargs = {
        "kind": kind,
        "env_setup": env_setup,
        "templates": templates,
        **param_args
    }

    if kind == EndpointKind.STREAM:
        kwargs["kafka_brokers"] = kafka_brokers
        kwargs["kafka_input_topics"] = kafka_input_topics

    endpoint = project.endpoints.create(title, file_name, function_name, **kwargs)

    success_message = messages.ENDPOINT_CREATE_SUCCESS.format(endpoint.title, endpoint.full_href)
    logger.log_and_echo(success_message)


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-n', '--name', prompt=messages.ENDPOINT_PROMPT_METRIC_NAME, help=messages.ENDPOINT_METRIC_NAME)
@click.option('-x', '--x', prompt=messages.ENDPOINT_PROMPT_METRIC_X, help=messages.ENDPOINT_METRIC_X)
@click.option('-y', '--y', prompt=messages.ENDPOINT_PROMPT_METRIC_Y, help=messages.ENDPOINT_METRIC_Y)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def log_metric(endpoint, logger, slug, name, y, x, project_slug):
    """
        logs a metric to the endpoint
    """
    if endpoint is None:
        logger.log_and_echo(messages.ENDPOINT_DOES_NOT_EXIST, error=True)

    endpoint.log_metric(name=name, y=y, x=x)


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def start(endpoint, logger, slug, project_slug):
    """
      starts the current endpoint
    """
    if endpoint is None:
        logger.log_and_echo(messages.ENDPOINT_DOES_NOT_EXIST, error=True)

    endpoint.start()
    logger.log_and_echo('Endpoint {} started!'.format(slug))


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def stop(endpoint, logger, slug, project_slug):
    """
      stop the current endpoint
    """
    if endpoint is None:
        logger.log_and_echo(messages.ENDPOINT_DOES_NOT_EXIST, error=True)

    endpoint.stop()
    logger.log_and_echo('Endpoint {} stopped!'.format(slug))


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-vs', '--version_slug', help=messages.ENDPOINT_HELP_VERSION_SLUG, required=True)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def rollback(endpoint, logger, slug, version_slug, project_slug):
    """
        Rollback an endpoint version
    """
    if endpoint is None:
        logger.log_and_echo(messages.Endpoint_DOES_NOT_EXIST, error=True)

    endpoint.rollback_version(version_slug=version_slug)

    logger.log_and_echo(messages.ENDPOINT_ROLLBACK_SUCCESS.format(slug, version_slug))


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-f', '--file_name', default=None, help=messages.ENDPOINT_HELP_FILE_NAME)
@click.option('-fn', '--function_name', default=None, help=messages.ENDPOINT_HELP_FUNCTION_NAME)
@click.option('-p', '--prep_file', default=None, help=messages.ENDPOINT_HELP_PREP_FILE)
@click.option('-pf', '--prep_function', default=None, help=messages.ENDPOINT_HELP_PREP_FUNCTION)
@click.option('-mt', '--max_timeout', default=None, help=messages.ENDPOINT_HELP_MAX_TIMEOUT)
@click.option('-a', '--args', multiple=True, default=None, help=messages.ENDPOINT_HELP_ARGS)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def update(endpoint,
           slug,
           logger,
           file_name,
           function_name,
           prep_file,
           prep_function,
           max_timeout,
           args,
           project_slug):
    """
        update an endpoint
    """

    if endpoint is None:
        logger.log_and_echo(messages.Endpoint_DOES_NOT_EXIST, error=True)

    param_args = {}
    if args:
        param_args = dict([p.split('=') for p in args])

    kwargs = {
        "prep_file": prep_file,
        "prep_function": prep_function,
        "max_timeout": max_timeout,
        "file_name": file_name,
        "function_name": function_name,
        **param_args
    }
    logger.log_and_echo(messages.ENDPOINT_UPDATE_IN_PROGRESS)
    endpoint = endpoint.update_version(**kwargs)

    success_message = messages.ENDPOINT_UPDATE_SUCCESS.format(endpoint.title)
    logger.log_and_echo(success_message)


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-l', '--logs', prompt=messages.ENDPOINT_PROMPT_LOG_MESSAGE,
              help=messages.ENDPOINT_HELP_LOG_MESSAGE)
@click.option('-lv', '--log_level', default=LOGS_TYPE_OUTPUT, help=messages.ENDPOINT_HELP_LOGS_TYPE)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def log(endpoint, logger, slug, logs, log_level, project_slug):
    """
        send log to the endpoint
    """
    split_logs = logs.split(",")

    if endpoint is None:
        logger.log_and_echo(messages.ENDPOINT_DOES_NOT_EXIST, error=True)

    if log_level not in (LOGS_TYPE_OUTPUT, LOGS_TYPE_ERROR, LOGS_TYPE_INFO, LOGS_TYPE_WARNING):
        logger.log_and_echo(messages.ENDPOINT_LOG_LEVEL_NOT_SUPPORTED.format(log_level), error=True)

    endpoint.log(logs=split_logs, log_level=log_level)


@endpoint_group.command()
@click.option('-s', '--slug', default=None, help=messages.ENDPOINT_SLUG)
@click.option('-st', '--start_time', default=None, help=messages.ENDPOINT_HELP_START_TIME)
@click.option('-et', '--end_time', default=None, help=messages.ENDPOINT_HELP_END_TIME)
@click.option('-o', '--offset', default=None, help=messages.ENDPOINT_HELP_OFFSET)
@click.option('-si', '--size', default=50, help=messages.ENDPOINT_HELP_BATCH_SIZE)
@click.option('-m', '--model', default=None, help=messages.ENDPOINT_HELP_MODEL_NUMBER)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def get_predictions(endpoint, logger, slug, start_time, end_time, offset, size, model, project_slug):
    """
        Get last 50 predictions of the endpoint
    """

    if endpoint is None:
        logger.log_and_echo(messages.ENDPOINT_DOES_NOT_EXIST, error=True)

    kwargs = {
        "size": size,
    }
    if start_time:
        kwargs['start_time'] = start_time
    if end_time:
        kwargs['end_time'] = end_time
    if offset:
        kwargs['offset'] = offset
    if model:
        kwargs['model'] = model

    res = endpoint.get_predictions(**kwargs)

    if res["total"] == 0:
        logger.log_and_echo(messages.ENDPOINT_NO_PREDICTIONS)
        return

    pretty_print_predictions(res["predictions"])



import click
from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.cli.utils.dict_param import DICT
from cnvrgv2.modules.datasource import StorageTypes


@click.group(name='datasource')
def datasource_group():
    pass


# TODO Rethink aws concrete parameters like bucket name and path

@datasource_group.command(name="create")
@click.option('-n', '--name', prompt=messages.DATASOURCE_NAME, help=messages.DATASOURCE_NAME_HELP)
@click.option('-t', '--type', type=click.Choice(StorageTypes), prompt=messages.DATASOURCE_TYPE,
              help=messages.DATASOURCE_TYPE_HELP)
@click.option('-bn', '--bucket-name', prompt=messages.DATASTORE_BUCKET, help=messages.DATASTORE_BUCKET_HELP)
@click.option('-c', '--credentials', type=DICT, prompt=messages.DATASOURCE_CREDENTIALS,
              help=messages.DATASOURCE_CREDENTIALS_HELP)
@click.option('-r', '--region', prompt=messages.DATASOURCE_REGION, help=messages.DATASOURCE_REGION_HELP)
@click.option('-d', '--description', prompt=messages.DATASOURCE_DESCRIPTION, help=messages.DATASOURCE_DESCRIPTION_HELP)
@click.option('-p', '--public', is_flag=True, prompt=messages.DATASOURCE_PUBLIC, help=messages.DATASOURCE_PUBLIC_HELP)
@click.option('-pa', '--path', prompt=messages.DATASOURCE_PATH, help=messages.DATASOURCE_PATH_HELP)
@click.option('-e', '--endpoint', prompt=messages.DATASOURCE_ENDPOINT, help=messages.DATASOURCE_ENDPOINT)
@click.option('-u', '--users', default=[], prompt=messages.DATASOURCE_USERS, help=messages.DATASOURCE_USERS_HELP)
@prepare_command()
def create_datasource(cnvrg, logger, name, storage_type, bucket_name, credentials, region, description, public, path,
                      endpoint, users):
    """
    @param users:
    @param endpoint:
    @param path:
    @param public:
    @param description:
    @param region:
    @param credentials:
    @param bucket_name:
    @param storage_type:
    @param name:
    @param cnvrg:
    @param logger:
    @param slug:
    @return:
    """
    try:
        ds = cnvrg.datasources.create(
            name=name,
            storage_type=storage_type,
            bucket_name=bucket_name,
            credentials=credentials,
            region=region,
            description=description,
            public=public,
            path=path,
            endpoint=endpoint,
            users=users
        )
        logger.log_and_echo(f"A new datasource with the slug: {ds.slug} created!")
    except Exception as e:
        logger.log_and_echo(f"Failed to create datasource. Error: {e}")


@datasource_group.command(name="delete-datasource")
@click.option('-s', '--slug', prompt=messages.DATASOURCE_SLUG_HELP, help=messages.DATASOURCE_SLUG)
@prepare_command()
def delete_datasource(cnvrg, logger, slug):
    """
    @param cnvrg:
    @param logger:
    @param slug:
    @return:
    """
    try:
        ds = cnvrg.datasources.get(slug)
        ds.delete()
    except Exception:
        logger.log_and_echo("Failed to delete datasource")


@datasource_group.command()
@click.option('-s', '--slug', prompt=messages.DATASOURCE_SLUG_HELP, help=messages.DATASOURCE_SLUG)
@click.option('-p', '--page-size', help=messages.DATASOURCE_PAGE_SIZE_HELP)
@click.option('-w', '--max-workers', help=messages.DATASOURCE_MAX_WORKERS_HELP)
@click.option('-f', "--force", is_flag=True, help=messages.DATASOURCE_FORCE)
@click.option('-i', '--skip-if-exists', is_flag=True, help=messages.DATASOURCE_IGNORE_FILE_ERROR)
@prepare_command()
def clone(cnvrg, logger, slug, page_size, max_workers, skip_if_exists, force):
    """
    @param cnvrg:
    @param logger:
    @param slug:
    @param page_size:
    @param max_workers:
    @param skip_if_exists:
    @param force:
    @return:
    """
    ds = cnvrg.datasources.get(slug)
    logger.log_and_echo('starting cloning datasource...')
    ds.clone(page_size=page_size, max_workers=max_workers, skip_if_exists=skip_if_exists, force=force)
    logger.log_and_echo('datasource cloned')


@datasource_group.command(name='list')
@click.option('-s', '--slug', prompt=messages.DATASOURCE_SLUG_HELP, help=messages.DATASOURCE_SLUG)
@prepare_command()
def list_command(cnvrg, slug):
    """
    @param cnvrg:
    @param slug:
    @return:
    """
    ds = cnvrg.datasources.get(slug)
    files = ds.list()
    table = PrettyTable()
    table.field_names = ["File name"]
    table.align["File name"] = 'l'
    for file in files:
        table.add_row([file])

    click.echo(table)



import pytest
import configparser
from tests.conftest import random_string

datasource_test_bukcet_name = 'cnvrg-datasources-bucket'
datasource_test_region = 'us-east-2'
unprocessable_entity_status_code = 422


def parse_aws_credentials(file_path):
    config = configparser.ConfigParser()
    config.read(file_path)

    credentials = {}
    for section in config.sections():
        credentials[section] = {
            'aws_access_key_id': config.get(section, 'aws_access_key_id'),
            'aws_secret_access_key': config.get(section, 'aws_secret_access_key')
        }
    return credentials


@pytest.fixture(scope="function")
def e2e_datasource(request, e2e_client):
    datasource_name = random_string(5)
    datasource = e2e_client.datasources.create(datasource_name)

    request.addfinalizer(lambda: datasource.delete())

    return datasource



from cnvrgv2.config import routes
from cnvrgv2.config.error_messages import LOCAL_COMMIT_DOESNT_EXIST_ERROR
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.base.data_owner import DataOwner
from cnvrgv2.modules.commits.project_commit import ProjectCommit
from cnvrgv2.modules.flows.flows_client import FlowsClient
from cnvrgv2.modules.labels.project_labels_client import ProjectLabelsClient
from cnvrgv2.modules.project_settings import ProjectSettings
from cnvrgv2.modules.volumes.volumes_client import VolumesClient
from cnvrgv2.modules.workflows import EndpointsClient, ExperimentsClient, WebappsClient, WorkspacesClient
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator


class Project(DataOwner):

    available_attributes = {
        "title": str,
        "p_tags": list,
        "git": bool,
        "start_commit": str,
        "commit": str,
        "git_url": str,
        "git_branch": str,
        "num_files": int,
        "last_commit": str,
        "public": bool
    }

    def __init__(self, context=None, slug=None, attributes=None):
        # Init data attributes
        super().__init__()

        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.PROJECT, slug)

        scope = self._context.get_scope(SCOPE.PROJECT)

        self._proxy = Proxy(context=self._context)
        self._route = routes.PROJECT_BASE.format(scope["organization"], scope["project"])
        self._attributes = attributes or {}
        self.slug = scope["project"]

        self._init_clients()

    def save(self):
        pass

    def delete(self):
        """
        Deletes the current project
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)

    def _init_clients(self):
        self.workspaces = WorkspacesClient(self)
        self.endpoints = EndpointsClient(self)
        self.webapps = WebappsClient(self)
        self.experiments = ExperimentsClient(self)
        self.volumes = VolumesClient(self)
        self.settings = ProjectSettings(self)
        self.flows = FlowsClient(self._context)
        self.labels = ProjectLabelsClient(self)

    def _validate_config_ownership(self):
        return self.slug == self._config.project_slug

    def save_config(self, local_config_path=None):
        """
        Saves the project configuration in the local folder
        @return: None
        """

        if not self.local_commit:
            raise CnvrgError(LOCAL_COMMIT_DOESNT_EXIST_ERROR)

        self._config.update(local_config_path=local_config_path, **{
            "project_slug": self.slug,
            "organization": self._context.organization,
            "git": getattr(self, "git", False),
            "commit_sha1": self.local_commit
        })

    def list_commits(self, filter={}, sort="-id"):
        """
        List all commits in a specific project
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields commit objects
        """
        scope = self._context.get_scope(SCOPE.PROJECT)
        list_commits_url = routes.PROJECT_COMMITS_BASE.format(scope["organization"], scope["project"])

        return api_list_generator(
            context=self._context,
            route=list_commits_url,
            filter=self._get_filter_object(filter),
            object=ProjectCommit,
            sort=sort,
            identifier="sha1"
        )

    def _get_filter_object(self, filter):

        conditions = [{
            'key': key if key != 'job_id' else 'job_slug',
            'operator': 'is',
            'value': filter[key]} for key in filter.keys()]

        return {
                "operator": 'AND',
                "conditions": conditions,
            }



import pytest
from tests.unit.modules.datasource.conftest import file_path


class TestDatasourceDownload:

    # Run the tests multiple times, with different params
    # Download with destination being relative or absolute path
    @pytest.mark.parametrize(
        "destination_path",
        [
            '/local-folder/my-file.txt',
            'local-folder/my-file.txt',
        ],
        ids=['1 - absolute destination path',
             '2 - relative destination path']
    )
    def test_download(self, datasource, mock_datasource_client, mock_normalize_file_path,
                      mock_datasource_files_downloader, destination_path):

        datasource.download_file(file_path, destination_path)
        datasource._normalize_file_path.assert_called_once_with(file_path)

        # Assert that the S3 client download_file method was called with correct arguments
        datasource.client.download_file.assert_called_once_with(
            Bucket=datasource.bucket_name,
            Key=mock_normalize_file_path.return_value,
            Filename=destination_path
        )



import pytest

from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgError
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.images import Image, ImageLogos


class MockImage:
    def __init__(self):
        self.slug = "test-slug"
        self.name = "test-image"
        self.tag = "v1"
        self.registry = "docker-hub"

        self.data = {
            "id": "17",
            "type": "project",
            "attributes": {
                "slug": self.slug,
                "name": self.name,
                "tag": self.tag,
                "dockerfile": "dockerfile",
                "registry_slug": self.registry,
                "logo": "",
                "readme": ""
            }
        }

    def response(self, *args, **kwargs):
        return JAF(response={"data": self.data})


class TestImage:
    def test_image_attributes(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        image = Image(empty_cnvrg, mock_resp.slug)
        assert image.slug == mock_resp.slug
        assert image.tag == mock_resp.tag
        assert image.name == mock_resp.name
        assert image.registry_slug == mock_resp.registry

        assert mocked_call.call_count == 1
        assert mocked_call.call_args[1]['http_method'] == 'GET'

    def test_delete_image(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        image = Image(empty_cnvrg, mock_resp.slug)
        image.delete()

        assert mocked_call.call_count == 1
        assert mocked_call.call_args[1]['http_method'] == 'DELETE'

    def test_image_dockerfile_property(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        image = Image(empty_cnvrg, mock_resp.slug)

        # Should only call the API once (property caching)
        image.dockerfile
        image.dockerfile
        image.dockerfile
        assert mocked_call.call_count == 1
        assert mocked_call.call_args[1]['http_method'] == 'GET'

    def test_update_image(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        image = Image(empty_cnvrg, mock_resp.slug)
        image.update(logo=ImageLogos.CNVRG)

        assert mocked_call.call_count == 1
        assert mocked_call.call_args[1]['http_method'] == 'PUT'

    def test_update_image_invalid_arguments(self, empty_context, empty_cnvrg):
        image = Image(empty_cnvrg, "slug")

        with pytest.raises(CnvrgError):
            image.update(logo="fake")



import shutil
import tempfile
from unittest.mock import patch
import pytest
from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgAlreadyClonedError
from cnvrgv2.modules.datasource.datasource_operations import handle_dir_exist


class TestHandleDirExist:
    @pytest.fixture(scope="function")
    def setup_directory(self):
        temp_dir = tempfile.mkdtemp()
        yield temp_dir
        shutil.rmtree(temp_dir)

    @patch('shutil.rmtree')
    def test_handle_dir_exist_force_true(self, mock_rmtree):
        path = 'test/path'
        handle_dir_exist(force=True, path=path)
        # Assert that shutil.rmtree was called with the correct path
        mock_rmtree.assert_called_once_with(path)

    @patch('shutil.rmtree')
    def test_handle_dir_exist_force_false(self, mock_rmtree):
        path = 'test/path'
        with pytest.raises(CnvrgAlreadyClonedError) as exception_info:
            handle_dir_exist(force=False, path=path)

        # Check if the exception message matches the expected error message
        assert error_messages.DATASOURCE_ALREADY_CLONED in str(exception_info.value)

    # this is e2e tests
    # def test_handle_dir_exist_force(self, setup_directory):
    #     # Create a subdirectory to test force deletion
    #     sub_dir = os.path.join(setup_directory, "test_sub_dir")
    #     os.mkdir(sub_dir)
    #     assert os.path.isdir(sub_dir)
    #
    #     # Call the function with force=True
    #     handle_dir_exist(force=True, path=sub_dir)
    #
    #     # Ensure the directory is deleted
    #     assert not os.path.exists(sub_dir)
    #
    # def test_handle_dir_exist_no_force(self, setup_directory):
    #     # Create a subdirectory to test non-force behavior
    #     sub_dir = os.path.join(setup_directory, "test_sub_dir")
    #     os.mkdir(sub_dir)
    #     assert os.path.isdir(sub_dir)
    #
    #     # Call the function with force=False and expect an exception
    #     with pytest.raises(CnvrgAlreadyClonedError) as excinfo:
    #         handle_dir_exist(force=False, path=sub_dir)
    #
    #     # Ensure the directory still exists
    #     assert os.path.isdir(sub_dir)
    #     assert str(excinfo.value) == "The datasource is already cloned"



from azure.storage.blob import BlobServiceClient

from cnvrgv2.data.clients.base_storage_client import BaseStorageClient
from cnvrgv2.utils.converters import convert_bytes
from cnvrgv2.utils.retry import retry
from cnvrgv2.utils.storage_utils import create_dir_if_not_exists


class AzureStorage(BaseStorageClient):
    def __init__(self, storage_meta):
        super().__init__(storage_meta)

        self.chunk_size_bytes = 4*1024*1024  # default chunk size for upload/download is 4MB

        props = self._decrypt_dict(storage_meta, keys=[
            "container",
            "storage_access_key",
            "storage_account_name",
        ])
        account_name = props["storage_account_name"]
        account_key = props["storage_access_key"]
        container = props["container"]

        self.access_key = (
            "DefaultEndpointsProtocol=https;"
            "AccountName={};"
            "AccountKey={};"
            "EndpointSuffix=core.windows.net"
        ).format(
            account_name,
            account_key
        )
        self.container_name = container
        self.service = self._get_service()

    @retry(log_error=True)
    def upload_single_file(self, local_path, object_path, progress_bar=None):

        try:
            client = self.service.get_blob_client(container=self.container_name, blob=object_path)

            # Init the blob
            if client.exists():
                client.delete_blob()
            client.create_append_blob()

            with open(local_path, "rb") as local_file:
                while True:
                    read_data = local_file.read(self.chunk_size_bytes)
                    if not read_data:
                        break
                    client.append_block(read_data)
                    self.update_progress(progress_bar, len(read_data))
        except Exception as e:
            print(e)

    @retry(log_error=True)
    def download_single_file(self, local_path, object_path, progress_bar=None):
        try:
            create_dir_if_not_exists(local_path)
            if not object_path:
                return

            client = self.service.get_blob_client(container=self.container_name, blob=object_path)

            with open(local_path, "wb") as local_file:
                stream = client.download_blob(max_concurrency=1)
                for chunk in stream.chunks():
                    local_file.write(chunk)
                    self.update_progress(progress_bar, len(chunk))
        except Exception as e:
            print(e)

    def update_progress(self, progress_bar, chunk_length_bytes):
        if progress_bar and progress_bar.max > 0:
            converted_bytes, _ = convert_bytes(chunk_length_bytes, unit=progress_bar.unit)
            with progress_bar.mutex:
                progress_bar.throttled_next(converted_bytes)

    def _get_service(self):
        return BlobServiceClient.from_connection_string(
            self.access_key,
            max_chunk_get_size=self.chunk_size_bytes,
            max_single_get_size=self.chunk_size_bytes
        )



from cnvrgv2.modules.workflows.workspace.workspace import Workspace


class TestWorkspace:
    def test_offline_workspace_restart_attribute(self, empty_context):
        try:
            w = Workspace(context=empty_context, slug="end-slug", attributes={'status': 'offline'})
            w.restart()
        except AttributeError as err:
            assert "'Workspace' object has no attribute 'restart'" in str(err)



from cnvrgv2.config import error_messages
from cnvrgv2.context import Context
from cnvrgv2.errors import CnvrgLoginError
from cnvrgv2.config import Config


def verify_login():
    # TODO: Prettify function, there is duplication between line 11 and line 13
    try:
        # Attempt to log in using credentials (from config files)
        credentials_variables = Config().get_credential_variables()
        # If cannot log in using config credentials, try using env variables
        auth_variables = credentials_variables if all(credentials_variables) else (Context().get_env_variables())
        if not all(auth_variables):
            raise CnvrgLoginError(error_messages.CREDENTIALS_MISSING)
    except CnvrgLoginError:
        raise CnvrgLoginError(error_messages.CREDENTIALS_MISSING)



from cnvrgv2.config import routes, error_messages
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.utils.validators import attributes_validator
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes


class Machine(DynamicAttributes):
    # TOOD: Add sshkey support
    available_attributes = {
        "slug": str,
        "title": str,
        "ssh_user": str,
        "ssh_ip": str,
        "ssh_port": str,
        "cpu_cores": float,
        "status": str,
        "ram": float,
        "gpu": bool,
        "gpu_cores": float,
        "gpu_model": str,
        "gpu_ram": float,
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current machine
        if slug:
            self._context.set_scope(SCOPE.RESOURCE, slug)

        scope = self._context.get_scope(SCOPE.RESOURCE)

        self._proxy = Proxy(context=self._context)
        self._route = routes.MACHINE_BASE.format(scope["organization"], scope["resource"])
        self._attributes = attributes or {}
        self.slug = scope["resource"]

    def update(self, **kwargs):
        """
        Updates given attributes of a machine

        @kwargs: A list of optional attributes to update
            title: Name of the new machine
            ssh_user: Credentials for the new machine
            ssh_ip: Host of the new machine
            ssh_port: Port of the new machine
            ssh_key: a ssh key file for the new machine(path to file relative to working dir)
            ssh_password: a password for the new machine
            cpu_cores: CPU cores for the new machine
            ram: Memroy for the new machine
            gpu: Does the new machine has a GPU
            gpu_model: What GPU the new machine has
            gpu_cores: GPU Cores for the new machine
            gpu_ram: GPU memroy for the new machine

        @return: the updated machine
        """

        # Validate attributes
        attributes_validator(self.available_attributes, kwargs)

        # Specific instance validations
        if (not kwargs.get('gpu') and not self._attributes['gpu'] and
                kwargs.get('gpu_model') or kwargs.get('gpu_cores') or kwargs.get('gpu_ram')):
            raise CnvrgArgumentsError(error_messages.MACHINE_GPU_VALUES_WITHOUT_GPU)

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type='machine', attributes={"slug": self.slug, **kwargs}))

        self._attributes = response.attributes

    def save(self):
        """
        In case of any attribute change, saves the changes
        """
        self.update(**self._attributes)

    def delete(self):
        """
        Deletes the current machine
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)



import pytest
from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgError, CnvrgArgumentsError
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.images import ImagesClient, Image, ImageLogos


class MockImage:
    def __init__(self, list=False):
        self.slug = "test-slug"
        self.name = "test-image"
        self.tag = "v1"
        self.registry = "docker-hub"
        self.data = {
            "id": "17",
            "type": "image",
            "attributes": {
                "slug": self.slug,
                "name": self.name,
                "tag": self.tag,
                "registry_slug": self.registry,
                "logo": "",
                "readme": ""
            }
        }
        if list:
            self.data = [self.data]

    def response(self, *args, **kwargs):
        return JAF(response={"data": self.data})


class TestImagesClient:
    def test_create_image_with_valid_arguments(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage()
        mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)
        client = ImagesClient(empty_cnvrg)
        image = client.create(
            name=mock_resp.name,
            tag=mock_resp.tag,
            registry=mock_resp.registry,
            logo=ImageLogos.CNVRG
        )
        assert isinstance(image, Image)
        assert image.slug == mock_resp.slug
        assert image.name == mock_resp.name

    def test_create_image_with_invalid_arguments(self, empty_context, empty_cnvrg):
        name = "test-image"
        tag = "v1"
        registry = "docker-hub"
        client = ImagesClient(empty_cnvrg)
        with pytest.raises(CnvrgError):
            client.create(name=name, tag=1, registry=registry, logo=ImageLogos.CNVRG)
        with pytest.raises(CnvrgError):
            client.create(name=name, tag="bad tag", registry=registry, logo=ImageLogos.CNVRG)
        with pytest.raises(CnvrgError):
            client.create(name=1, tag=tag, registry=registry, logo=ImageLogos.CNVRG)
        with pytest.raises(CnvrgError):
            client.create(name="bad name", tag=tag, registry=registry, logo=ImageLogos.CNVRG)
        with pytest.raises(CnvrgError):
            client.create(name=name, tag=tag, registry=registry, logo="fake")

    def test_list_images(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage(list=True)
        mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)
        client = ImagesClient(empty_cnvrg)
        images = client.list()
        images = [image for image in images]
        assert len(images) == 1
        assert isinstance(images[0], Image)
        assert images[0].slug == mock_resp.slug
        assert images[0].name == mock_resp.name

    def test_get_image(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage()
        mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)
        client = ImagesClient(empty_cnvrg)
        image = client.get(mock_resp.slug)
        assert isinstance(image, Image)
        assert image.slug == mock_resp.slug
        assert image.name == mock_resp.name

    def test_delete_image(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockImage()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)
        client = ImagesClient(empty_cnvrg)
        client.delete(mock_resp.slug)
        assert mocked_call.call_count == 1

    def test_create_image_with_valid_name(self, mocker, empty_context, empty_cnvrg):
        client = ImagesClient(empty_cnvrg)
        valid_names = ["test", "test-image", "test/image"]
        for name in valid_names:
            mock_data = {
                "id": "17",
                "type": "image",
                "attributes": {
                    "slug": "test-slug",
                    "name": name,
                    "tag": "v1",
                    "registry_slug": "docker-hub",
                    "logo": "",
                    "readme": ""
                }
            }
            mock_response = JAF(response={"data": mock_data})
            mocker.patch.object(Proxy, 'call_api', return_value=mock_response)
            image = client.create(name=name, tag="v1", registry="docker-hub")
            assert image.name == name
            assert image.tag == "v1"

    def test_create_image_with_invalid_name(self, mocker, empty_context, empty_cnvrg):
        invalid_names = ["test\\image", "test!%#", "test*invalid"]
        client = ImagesClient(empty_cnvrg)
        mock_resp = MockImage()
        for name in invalid_names:
            mocker.patch.object(Proxy, 'call_api', side_effect=CnvrgArgumentsError("Bad arguments: Image name is invalid"))
            with pytest.raises(CnvrgArgumentsError, match="Bad arguments: Image name is invalid"):
                client.create(name=name, tag=mock_resp.tag, registry=mock_resp.registry)



import pytest

from cnvrgv2.errors import CnvrgHttpError
from cnvrgv2.modules.organization.organizations_client import OrganizationsClient
from tests.e2e.conftest import call_database


class TestOrganizations:
    @staticmethod
    def cleanup(context_prefix):
        # TODO: this doesnt cleanup all of the organization clusters and stuff.
        delete_user_command = "DELETE FROM organizations WHERE slug LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    def test_create_organization(self, class_context, domain, e2e_user):
        org_name = class_context.generate_name(5)
        oc = OrganizationsClient(domain=domain, token=e2e_user["token"])
        sample_org = oc.create(name=org_name)
        assert sample_org["slug"] == org_name

    def test_create_duplicate_organization(self, class_context, domain, e2e_user):
        org_name = class_context.generate_name(5)
        oc = OrganizationsClient(domain=domain, token=e2e_user["token"])
        sample_org = oc.create(name=org_name)
        assert sample_org["slug"] == org_name

        with pytest.raises(CnvrgHttpError) as exception_info:
            oc.create(name=org_name)

        assert "exists" in str(exception_info.value)

    def test_create_organization_bad_name(self, class_context, domain, e2e_user):
        org_name = class_context.generate_name(5) + "-bad"
        oc = OrganizationsClient(domain=domain, token=e2e_user["token"])

        with pytest.raises(CnvrgHttpError) as exception_info:
            oc.create(name=org_name)

        assert "invalid title" in str(exception_info.value)

    def test_get_organization(self, class_context, domain, e2e_user):
        org_name = class_context.generate_name(5)
        oc = OrganizationsClient(domain=domain, token=e2e_user["token"])
        oc.create(name=org_name)

        sample_org = oc.get(slug=org_name)

        assert sample_org["slug"] == org_name

    def test_get_non_existent_organization(self, class_context, domain, e2e_user):
        org_name = class_context.generate_name(5)
        oc = OrganizationsClient(domain=domain, token=e2e_user["token"])

        with pytest.raises(CnvrgHttpError) as exception_info:
            oc.get(slug=org_name)

        assert "found" in str(exception_info.value)



from cnvrgv2.modules.workflows.webapp.webapp import Webapp, WebappType
from cnvrgv2.modules.workflows.experiment.experiment import Experiment
from cnvrgv2.modules.workflows.workspace.workspace import Workspace, NotebookType
from cnvrgv2.modules.workflows.endpoint.endpoint import Endpoint, EndpointKind, EndpointEnvSetup

from cnvrgv2.modules.workflows.webapp.webapps_client import WebappsClient
from cnvrgv2.modules.workflows.endpoint.endpoints_client import EndpointsClient
from cnvrgv2.modules.workflows.workspace.workspaces_client import WorkspacesClient
from cnvrgv2.modules.workflows.experiment.experiments_client import ExperimentsClient




import os
import sys

import click
import yaml
from prettytable import PrettyTable

from cnvrgv2.utils.url_utils import urljoin


def parse_parameters_from_file(path):
    """
    Parse parameters from command line
    :param parameters: parameters yaml
    :return: parameters list of dicts
    """
    if not os.path.exists(path):
        raise ValueError("The provided parameters yaml path does not exist.")

    with open(path, "r") as f:
        data = yaml.safe_load(f)

    if not data.get("parameters"):
        raise ValueError("The provided parameters yaml is missing the parameters key")

    return data.get("parameters")


def build_grid_url(cnvrg, project, grid_slug):
    """
    Build Grid url
    :param cnvrg: cnvrg object
    :param project: project object
    :param grid_slug: grid slug
    :return: grid url
    """
    grid_path = "{}/experiments?grid={}".format(project._route, grid_slug)
    route = urljoin(cnvrg._proxy._domain, "api", grid_path)
    return route


def callback_log(experiment, log=False):
    if not log:
        return

    end_pos = 0

    def callback():
        nonlocal end_pos
        nonlocal log
        try:
            resp = experiment.info(end_pos=end_pos)
            end_pos = resp.attributes["info"][experiment.slug]["end_pos"]
            logs = resp.attributes["info"][experiment.slug]["logs"]
            for log in logs:
                click.echo(log["message"])
        except Exception as e:
            click.echo("Failed to get logs, error: {}".format(e))

    return callback


def print_generator_in_chunks(generator, chunk_size, limit, object_attributes, attributes_field_name, line_numbers=False):
    table = PrettyTable()

    table.field_names = ["#"] + attributes_field_name if line_numbers else attributes_field_name
    for field in table.field_names:
        table.align[field] = "c"

    counter = 0
    limit = limit or sys.maxsize
    while True:
        chunk = []
        limit_reached = False
        for i in range(chunk_size):
            try:
                obj = next(generator)
                row = [obj.__getattr__(attribute) for attribute in object_attributes]
                if line_numbers:
                    row = [counter + 1] + row
                chunk.append(row)
                counter += 1
                if counter >= limit:
                    limit_reached = True
                    break
            except StopIteration:
                break

        if not chunk:
            # If num of items divides by chunk_size, this break will prevent printing an empty table at the end
            break

        # Clear rows of previous iteration and add rows of current iteration
        table.clear_rows()
        table.add_rows(chunk)

        # Clearing previous print will make it look like the content of the table changes but the structure remains
        click.echo("\n")
        click.echo(table)

        if limit_reached or len(chunk) != chunk_size:
            # Partial chunk, means it's the last one. break out of the while loop
            break
        else:
            click.echo("Press any key to show the next chunk of files...", nl=False)
            click.getchar()


def pretty_print_predictions(predictions):
    """
    Pretty print predictions of endpoint
    :@predictions: list of predictions dict
    :return: None
    """
    click.echo("{:<5} {:<20} {:<20} {:<20} {:<13}".format('Model',
                                                          'Time',
                                                          'Input',
                                                          'Output',
                                                          'Elapsed_time'))
    for prediction in predictions:
        click.echo("{:<5} {:<20} {:<20} {:<20} {:<13}".format(prediction["model"],
                                                              prediction["start_time"],
                                                              prediction["input"][:20],
                                                              prediction["output"][:20],
                                                              prediction["elapsed_time"]))



import pytest

from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.users.user import User
from cnvrgv2.proxy import Proxy
from cnvrgv2.utils.json_api_format import JAF


class TestUser:

    def test_user_init(self):
        name = "Israel Israeli"
        bio = "Working at cnvrg"
        email = "israeli.israeli@walla.co.il"
        attr = {
            "name": name,
            "bio": bio,
            "email": email
        }
        user = User(domain="domain", token="token", attributes=attr)

        assert user.name == name
        assert user.bio == bio

    def test_user_valid_update_success(self, mocker):
        username = "israeli123"
        name = "Israel Israeli"
        company = "cnvrg.io"
        bio = "A fullstack developer"
        time_zone = "Asia/Jerusalem"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "1",
                    "type": "user",
                    "attributes": {
                        "username": username,
                        "name": name,
                        "company": company,
                        "bio": bio,
                        "time_zone": time_zone
                    }
                }
            }

            return JAF(response=response)

        mock_response = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        user = User(domain="domain", token="token")
        user.update(
            **{"name": name,
               "company": company,
               "bio": bio,
               "time_zone": time_zone})

        assert mock_response.call_count == 1
        assert user.name == name
        assert user.bio == bio
        assert user.company == company
        assert user.time_zone == time_zone

    def test_user_save_success(self, mocker):
        username = "israeli123"
        name = "Israel Israeli"
        company = "cnvrg.io"
        bio = "A fullstack developer "
        time_zone = "Asia/Jerusalem"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "project",
                    "attributes": {
                        "username": username,
                        "name": name,
                        "company": company,
                        "bio": bio,
                        "time_zone": time_zone
                    }
                }
            }

            return JAF(response=response)

        mock_response = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        user = User(domain="domain", token="token")
        user.name = name
        user.bio = bio
        user.company = company
        user.time_zone = time_zone
        user.save()

        assert mock_response.call_count == 1
        assert user.name == name
        assert user.bio == bio
        assert user.company == company
        assert user.time_zone == time_zone

    def test_user_update_fake_icon_fail(self):
        file_path = "REALLY/FAKE/PATH.file"
        user = User(domain="domain", token="token",
                    attributes={
                        "icon": "https://fakeimg.com/img.png",
                        "name": "Israel Israeli",
                        "email": "israeli.israeli@walla.co.il"
                    })

        with pytest.raises(CnvrgArgumentsError):
            user.update(**{"icon": file_path})

    def test_user_update_fake_vscode_settings_fail(self):
        file_path = "REALLY/FAKE/PATH.file"
        user = User(
            domain="domain",
            token="token",
            attributes={
                "icon": "https://fakeimg.com/img.png",
                "name": "Israel Israeli",
                "email": "israeli.israeli@walla.co.il"
            }
        )

        with pytest.raises(CnvrgArgumentsError):
            user.update(**{"vscode_settings": file_path})

    def test_user_leave_org_success(self, mocker):
        # Void leave org.
        mock_leave = mocker.patch.object(Proxy, "call_api")
        user = User(domain="Domain", token="Token", attributes={
            "organizations": [
                {"id": 1, "slug": "MY_ORG"},
                {"id": 2, "slug": "MY_OTHER_ORG"}
            ],
            "email": "israeli.israeli@walla.co.il"
        },
                    )
        org = user.organizations[0]  # MY_ORG

        user.leave_org(org["slug"])

        assert mock_leave.call_count == 1
        assert org not in user.organizations



import pytest

from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.utils.chart_utils import BarChart, CHART_TYPES, Heatmap, LineChart, ScatterPlot


class TestCharts:
    def test_create_line_chart(self):
        new_chart = LineChart(key="test_chart")

        assert new_chart.chart_type == CHART_TYPES.LINE

    def test_create_bar_chart(self):
        new_chart = BarChart(key="test_bar")

        assert new_chart.chart_type == CHART_TYPES.BAR

    def test_create_scatter_chart(self):
        new_chart = ScatterPlot(key="test_scatter")

        assert new_chart.chart_type == CHART_TYPES.SCATTER

    def test_create_heatmap_chart(self):
        new_chart = Heatmap(key="test_heatmap")

        assert new_chart.chart_type == CHART_TYPES.HEATMAP

    def test_add_regular_series(self):
        new_chart = LineChart(key="test_chart")
        series_name = "test"
        series_values = [1, 2, 3]

        new_chart.add_series(series_values, series_name)
        assert new_chart.series_list[0].get("data")[0] == 1

    def test_add_faulty_regular_series(self):
        new_chart = LineChart(key="test_faulty_series")
        series_name = "test"
        series_values = 'a'

        with pytest.raises(CnvrgArgumentsError) as exception_info:
            new_chart.add_series(series_name, series_values)

        assert "Series" in str(exception_info.value)

    def test_add_non_numeric_series(self):
        new_chart = LineChart(key="test_faulty_series")
        series_name = "test"
        series_values = [1, 'a', 3]

        with pytest.raises(CnvrgArgumentsError) as exception_info:
            new_chart.add_series(series_name, series_values)

        assert "Series" in str(exception_info.value)

    def test_add_heatmap_series(self):
        new_chart = Heatmap(key="test_chart")
        series_name = "test"
        series_values = [(1, 2, 3), (1, 2, 3)]

        new_chart.add_series(series_values, series_name)
        assert new_chart.series_list[0].get("data")[0] == (0, 0, 1)

    def test_add_faulty_heatmap_value_length(self):
        new_chart = Heatmap(key="test_heatmap_faulty")
        series_name = "test"
        series_values = [(1, 2), (1, 2, 3), (1, 2)]

        with pytest.raises(CnvrgArgumentsError) as exception_info:
            new_chart.add_series(series_values, series_name)

        assert "Heatmap" in str(exception_info.value)

    def test_add_faulty_heatmap_value_type(self):
        new_chart = Heatmap(key="test_heatmap_faulty")
        series_name = "test"
        series_values = [(1, 2, 3), (1, 2, 'a'), (1, 2, 3)]

        with pytest.raises(CnvrgArgumentsError) as exception_info:
            new_chart.add_series(series_values, series_name)

        assert "Heatmap" in str(exception_info.value)

    def test_series_add_same_name_multiple_times(self):
        new_chart = LineChart(key="test")
        new_chart.add_series([1, 2, 3], name="test")

        try:
            new_chart.add_series([4, 5, 6], name="test")
        except AttributeError:
            raise pytest.fail("DID RAISE {0}".format(AttributeError))

    def test_add_faulty_heatmap_step(self):
        with pytest.raises(CnvrgArgumentsError) as exception_info:
            Heatmap(key="test_heatmap_faulty", step=1)

        assert "group" in str(exception_info.value)

    def test_add_group_and_step_non_heatmap(self):
        with pytest.raises(CnvrgArgumentsError) as exception_info:
            LineChart(key="non_hm_group", group="test", step=1)

        assert "Heatmaps" in str(exception_info.value)



import os
import shutil
import time
from datetime import datetime

from cnvrgv2.config import routes, error_messages
from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.base.workflow_instance_base import WorkflowInstanceBase
from cnvrgv2.modules.workflows.workflow_utils import WorkflowStatuses
from cnvrgv2.proxy import Proxy
from cnvrgv2.context import Context, SCOPE


class WebappType:
    SHINY = "rshiny"
    DASH = "dash"
    VOILA = "voila"
    TENSORBOARD = "tensorboard"


class Webapp(WorkflowInstanceBase):
    available_attributes = {
        "webapp_type": str,
        "template_ids": list,
        "template_id": int,
        "compute": str,
        "num_of_exps": int,
        "updated_at": datetime,
        "iframe_url": str,
        "is_public": bool,
        "last_opened": datetime,
        "current_step": int,
        "strip_sources": bool,
        "copy_frequency": int,
        "file_name": str,
        "experiments": list,
        **WorkflowInstanceBase.available_attributes
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.WEBAPP, slug)

        self.scope = self._context.get_scope(SCOPE.WEBAPP)

        self._proxy = Proxy(context=self._context)
        self._route = routes.WEBAPP_BASE.format(self.scope["organization"], self.scope["project"], self.scope["webapp"])
        self._attributes = attributes or {}
        self._type = "Webapp"
        self.slug = self.scope["webapp"]

    def start(self):
        """
        Override start from workflows_base to remove functionality.
        start() is only relevant for Endpoints & Workspaces
        """
        raise AttributeError("'Webapp' object has no attribute 'start'")

    def sync_remote(self, commit_msg=None):
        raise NotImplementedError()

    def compare_experiments(self, frequency=5):
        """
        Used for tensorboard, Download tfevents to predefined location {exp_slug}/{project}/{files}
        New experiments can be added to the webapp experiments attribute in server side.
        @param frequency: [integer] number in seconds used for interval between each commit download
        """
        # Importing here to avoid circular dependencies (webapp <-> project)
        from cnvrgv2 import Project
        project = Project(context=self._context, slug=self.scope["project"])

        exps_map = {}

        while True:
            self.reload()
            slugs = self.experiments
            if not slugs:
                raise CnvrgError(error_messages.EXPERIMENT_SLUGS_NOT_FOUND)

            for slug in slugs:
                if slug not in exps_map:
                    exp = project.experiments.get(slug)
                else:
                    # end commit already cloned, skipping it
                    continue
                if exp.end_commit is not None and exp.status != WorkflowStatuses.ONGOING:
                    self._download_tfevents_from_experiment(exp, project, exp.end_commit)
                    exps_map[slug] = exp
                else:
                    self._download_tfevents_from_experiment(exp, project, exp.last_successful_commit)
            time.sleep(frequency)

    def _download_tfevents_from_experiment(self, exp, project, commit_sha1):
        """
        Download tensorflow events to predefined location {exp_slug}/{project}/{files} using %tfevents% filter
        New experiments can be added to the webapp experiments attribute in server side.
        @param exp: [Experiment] Experiment object
        @param project: [Project] Project object
        @param commit_sha1: [String] the sha1 of the commit to download
        """
        # Check if commit_sh1 exists
        if commit_sha1 is None:
            return

        # Set destination folder
        base_dir = exp.slug

        # Clear previous logs
        if os.path.isdir(base_dir):
            shutil.rmtree(base_dir)

        # Change working_dir to destination folder
        old_working_dir = project.working_dir
        project.working_dir = base_dir

        # Download the commit with the requested filter
        project.clone(commit=commit_sha1, fullpath_filter="%tfevents%")

        # Revert to the previous working_dir
        project.working_dir = old_working_dir




class JAF:
    def __init__(self, response):
        # Parse meta
        self.meta = response.get("meta", {})

        # Parse data
        data = response.get('data', {})
        if isinstance(data, dict):
            self._parse_from_object(data)
        elif type(data) == list:
            self.items = []
            for item in data:
                self.items.append(JAF({"data": item}))

        # Parse cursor from meta
        self.next = self.meta.get('next', '')

    @staticmethod
    def serialize(type, attributes, relationships=None):
        """
        Creates a JSON API valid request
        @param type: The type of object we want to modify
        @param attributes: The objects attributes
        @param relationships: Any relationships should be passed through this filed
        @return:
        """
        data_obj = {
            "type": type,
            "attributes": attributes,
        }

        if relationships:
            data_obj["relationships"] = relationships

        return {"data": data_obj}

    def _parse_from_object(self, data):
        """
        Parses the data json into self.
        @param data: json in Json API Specification format
        @return: None
        """
        self.id = data.get("id", None)
        self.type = data.get("type", None)
        self.attributes = data.get("attributes", {})

        self.relationships = data.get("relationships", {})
        for key in self.relationships:
            self.relationships[key] = JAF(self.relationships.get(key, {}))



from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config.config import SSO_VERSION
from cnvrgv2.config import routes, error_messages
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import Proxy, HTTP
import os

from cnvrgv2.utils.url_utils import encode_base64


class ROLES:
    ADMIN = 'admin'
    MANAGER = 'manager'
    MEMBER = 'member'
    REVIEWER = 'reviewer'
    ALL_ROLES = [ADMIN, MANAGER, MEMBER, REVIEWER]


class User(DynamicAttributes):
    available_attributes = {
        "icon": str,
        "username": str,
        "name": str,
        "bio": str,
        "company": str,
        "time_zone": str,
        "git_access_token": str,
        "default_homepage": str,
        "vscode_settings": str,
        "roles": list,
        "organizations": list
    }

    def __init__(self, domain, token, sso_version=SSO_VERSION.CAPI, attributes=None):
        self._proxy = Proxy(domain=domain, token=token, sso_version=sso_version)
        if attributes:
            self._email = attributes["email"]
            attributes["icon"] = None
            attributes["vscode_settings"] = None

        self._token = token
        self._route = routes.USER_CURRENT
        self._attributes = attributes or {}

    @property
    def email(self):
        return self._email

    def save(self):
        """
        In case of any attribute change, saves the changes
        """
        self.update(**self._attributes)

    def update(self, **kwargs):
        """
        Update the users profile
        @param kwargs: icon, username, name, password, bio, company, time_zone,
        git_access_token, default homepage, vscode_settings
        """
        if kwargs.get("icon") and not os.path.exists(kwargs.get("icon")):
            raise CnvrgArgumentsError(error_messages.NO_FILE)

        if kwargs.get("vscode_settings") and not os.path.exists(kwargs.get("vscode_settings")):
            raise CnvrgArgumentsError(error_messages.NO_FILE)

        kwargs["icon"] = encode_base64(kwargs.get("icon"))
        kwargs["vscode_settings"] = encode_base64(kwargs.get("vscode_settings"))

        response = self._proxy.call_api(route=routes.USER_CURRENT, http_method=HTTP.PUT, payload=kwargs)
        self._attributes = response.attributes

    def leave_org(self, slug):
        """
        leave organization - remove it from list of organizations and update database
        @param slug: organization slug
        """
        if not any(org.get("slug") == slug for org in self._attributes["organizations"]):
            raise CnvrgArgumentsError(error_messages.ORGANIZATION_NOT_FOUND)

        self._proxy.call_api(route=routes.USER_MEMBERSHIPS + slug, http_method=HTTP.DELETE)
        self._attributes["organizations"] = list(
            filter(lambda org: org.get("slug") != slug, self._attributes["organizations"])
        )



from cnvrgv2 import Project
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.workflows.experiment.experiment import Experiment


class TestExperiment:
    def test_stopped_experiment_restart_attribute(self, empty_context):
        try:
            e = Experiment(context=empty_context, slug="end-slug", attributes={'status': 'stopped'})
            e.restart()
        except AttributeError as err:
            assert "'Experiment' object has no attribute 'restart'" in str(err)

    def test_logs_experiment_offset_attribute(self, empty_context):
        try:
            e = Experiment(context=empty_context, slug="end-slug", attributes={'status': 'stopped'})
            e.logs(offset="test")
        except CnvrgArgumentsError as err:
            assert "offset: Wrong attribute type, expected int, got <class 'str'>" in str(err)

    def test_logs_experiment_page_attribute(self, empty_context):
        try:
            e = Experiment(context=empty_context, slug="end-slug", attributes={'status': 'stopped'})
            e.logs(page="test")
        except CnvrgArgumentsError as err:
            assert "page: Wrong attribute type, expected dict, got <class 'str'>" in str(err)

    def test_create_experiment_parameters_duplicate_key_error(self, empty_context):
        try:
            params = [
                {
                    "key": "",
                    "type": "discrete",
                    "min": 0,
                    "max": 0,
                    "scale": "linear",
                    "steps": 0,
                    "values": [
                        1,
                        2,
                        3,
                    ]
                },
                {
                    "key": "key2",
                    "type": "float",
                    "min": 1,
                    "max": 40,
                    "scale": "linear",
                    "steps": 3,
                    "values": []
                },
            ]
            project = Project(context=empty_context, slug="test-slug")
            project.experiments.create_grid(title="test", command="echo 'test'", parameters=params, sync_before=False)
        except CnvrgArgumentsError as err:
            assert "Key cannot be empty" in str(err)

    def test_create_experiment_parameters_discrete_values_error(self, empty_context):
        try:
            params = [
                {
                    "key": "key1",
                    "type": "discrete",
                    "min": 0,
                    "max": 0,
                    "scale": "linear",
                    "steps": 0,
                    "values": [
                        "1",
                        2,
                        3,
                    ]
                },
                {
                    "key": "key2",
                    "type": "float",
                    "min": 1,
                    "max": 40,
                    "scale": "linear",
                    "steps": 3,
                    "values": []
                },
            ]
            project = Project(context=empty_context, slug="test-slug")
            project.experiments.create_grid(title="test", command="echo 'test'", parameters=params, sync_before=False)
        except CnvrgArgumentsError as err:
            assert "Discrete values may only contain numbers" in str(err)

    def test_create_experiment_parameters_command_missing_error(self, empty_context):
        try:
            params = [
                {
                    "key": "key1",
                    "type": "discrete",
                    "min": 0,
                    "max": 0,
                    "scale": "linear",
                    "steps": 0,
                    "values": [
                        1,
                        2,
                        3,
                    ]
                },
                {
                    "key": "key2",
                    "type": "float",
                    "min": 1,
                    "max": 40,
                    "scale": "linear",
                    "steps": 3,
                    "values": []
                },
            ]
            project = Project(context=empty_context, slug="test-slug")
            project.experiments.create_grid(title="test", parameters=params, sync_before=False)
        except CnvrgArgumentsError as err:
            assert "Command Cannot be blank" in str(err)
